/*----------------------------------------------------------------------*
 * List.h
 *
 * Michael Geikowski, 5.3.93
 *
 * Definiert die Klasse List und deren Hilfsklassen ListElem und 
 * ListLink. Die Klasse List beschreibt eine einfach verkettete Liste
 * von Elementen vom Typ ListElem oder davon abgeleiteten Typen.
 *----------------------------------------------------------------------*/

#ifndef LIST_H
#define LIST_H

#include <assert.h>

class ListElem
//============
	// ListElem bildet die Basisklasse aller Listenelemente beschreibender
        // Klassen. Ein Listenelement vom Typ ListElem oder eines davon
        // abgeleiteten Typs kann in beliebig viele Listen eingehaengt werden.
        // Wird eine Liste, in der ein Element enthalten ist, geloescht, so
        // wird das Element nur dann ebenfalls geloescht, wenn es in keiner
        // weiteren Liste mehr enthalten ist.
 {
 friend class ListLink;
 friend class List;

 public:

   ListElem()
 //----------
	:count(0)
	{}
   virtual ~ListElem()
 //-------------------
	{}

 private:

   int count;			// Zaehler, wie oft das Element in Listen
				// eingehaengt wurde
 };


class ListLink
//============
	// ListLink ist eine Hilfsklasse fuer die Klasse List.
 {
 friend class List;

 public:

   ListLink(ListElem* e);
 //----------------------
   	// Neuer Link zeigt auf Element e
        
   ~ListLink();
 //------------

 private:

    ListElem *elem;		// ein Link zeigt auf ein Element und
    ListLink *next;		// einen weiteren Link (oder NULL)

 };

 
class List
//========
	// Die Klasse List beschreibt eine einfach verkettete Liste mit einer
        // Reihe von Zugriffsoperationen. Eine Liste dieses Typs ist aufgebaut
        // aus verketteten Links vom Typ ListLink. Jeder Link zeigt auf ein
        // Element vom Typ ListElem.
 {
 public:

   List (ListElem *e = 0);
 //-----------------------
	// Erzeuge leere (e==0) oder einelementige (e!=0) Liste.

   ~List ();
 //---------
	// Alle Links der Liste und alle Elemente, die in keiner anderen Liste
        // mehr eingetragen sind, werden geloescht.

   void Append (List *l);
 //----------------------
	// Haenge Liste l hinten an Liste an.
	// !! l wird dabei geloescht !!

   void Append (ListElem *e);
 //--------------------------
	// Haenge Element e hinten an Liste an.

   ListElem* GetFirstElem();
 //-------------------------
	// Liefert erstes Element der Liste. Null, wenn Liste leer.

   ListElem* GetLastElem();
 //------------------------
	// Liefert letztes Element der Liste. Liste darf nicht leer sein, sonst
        // Programmabbruch.

   ListElem* GetNextElem();
 //------------------------
	// Nach Aufruf von GetFirstElem kann die Liste mit GetNextElem
        // durchlaufen werden. Es wird so lange ein weiteres Element geliefert,
        // bis das Listenende erreicht wurde. Anschliessend wird Null geliefert.

   ListElem* GetCurrentElem();
 //---------------------------
	// Liefert das letzte Element des aktuellen Durchlaufs noch einmal.

   void SetCurrentElem(ListElem *e);
 //---------------------------------
	// Tauscht das letzte Element des aktuellen Durchlaufs gegen das
        // Element e aus. Ist das letzte Element in keiner anderen Liste
        // mehr enthalten, wird es geloescht. Programmabbruch, wenn kein
        // aktuelles Element existiert.

   void DelCurrentElem();
 //----------------------
	// Das letzte Element des aktuellen Durchlaufs wird aus der Liste
        // entfernt und geloescht, wenn es in keiner anderen Liste mehr
        // vorkommt. Programmabbruch, wenn kein aktuelles Element existiert.

   void DelFirstElem();
 //--------------------
	// Das erste Element (mit zugehoerigem Link) wird aus der Liste
        // entfernt. Ist das Element in keiner anderen Liste mehr enthalten,
        // wird es geloescht. Programmabbruch, wenn Liste leer.

   void NewFirstElem(ListElem *e);
 //-------------------------------
	// Haenge Element e vorne an Liste an.

   int is_Singleton();
 //-------------------
	// true(!=0), wenn Liste einelementig.

   int GetLength();
 //----------------
	// Liefert Listenlaenge.

 private:

   ListLink *first;		// erster Link
   ListLink *last;		// letzter Link
   ListLink *current;		// aktueller Link im aktuellen Durchlauf
				// ( Null, wenn Durchlauf beendet ).
   int length;			// Listenlaenge
 };

#endif
