/*---------------------------------------------------------------------------*
 * NavWin.c
 *
 * Michael Geikowski, 21.6.93
 *
 * Generierung und Verwaltung des Navigation und des Bookmark Windows
 *---------------------------------------------------------------------------*/

#include <assert.h>

#include "navwin.h"
#include "modwin.h"
#include "link.h"
#include "dbinterface.h"

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/MwmUtil.h>
#include <Xm/PushB.h>
#include <Xm/SelectioB.h>
#include <Xm/Separator.h>
#include <Xm/Text.h>
#include <X11/Shell.h>

// Button Identifier
//------------------

#define NAVDEF    1
#define NAVPREV   2
#define NAVNEXT   3
#define NAVMOD    4
#define BMGOTO   10
#define BMADD    11
#define BMDELETE 12
#define BMCANCEL 13
#define PDOK     20
#define PDCANCEL 21

// Modulglobale Variablen
//-----------------------

static Link* lastlink=NULL;	// letzter Angezeigter Link
static char* lastmod;		// Modul, das lastlink enthaelt
static Widget navshell=NULL;	// global sichtbare Widgets
static Widget label;
static Widget button1,button2,button3,button4;
static Widget bookmarkbox=NULL;
static Widget AddButton,DeleteButton,GotoButton,PromptDialog;
static long*  BookmarkDestPos;	// Sprungziele der Lesezeichen
static char** BookmarkDestName;
static int    BookmarkCount;	// Anzahl Lesezeichen

// Callbacks
//----------

static void NavigateCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Ein Button des NavWins wurde aktiviert
 {
 int button = (int) client_data;
 char* Mod;
 long pos;

 assert(lastlink);
 switch (button)
   {
   case NAVDEF:		// zur Definition
     lastlink->GetDef(&pos,&Mod);
     BeamToPosAtMod(pos,Mod);
     break;
   case NAVPREV:	// zum letzten Auftreten
     lastlink->GetPrevUse(&pos,&Mod);
     BeamToPosAtMod(pos,Mod);
     break;
   case NAVNEXT:	// zum neaechsten Auftreten
     lastlink->GetNextUse(&pos,&Mod);
     BeamToPosAtMod(pos,Mod);
     break;
   case NAVMOD:		// zu importiertem Modul
     lastlink->GetMod(&pos,&Mod);
     BeamToPosAtMod(pos,Mod);
     break;
   }
 }

static void BookmarkCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Button des Bookmarkwindows wurde aktiviert
 {
 int button = (int) client_data;
 Widget list,text;
 Boolean mem;
 int *pos_list;
 int pos_count,i;
 char* DestMod;
 long  DestPos;
 Arg  al[10];
 int  ac;
 XmString value_str;

 switch (button)
   {
   case BMGOTO:		// Folge dem Lesezeichen
     assert(BookmarkCount>0);
     list=XmSelectionBoxGetChild(bookmarkbox,XmDIALOG_LIST); // hole Nr. in
     mem=XmListGetSelectedPos(list,&pos_list,&pos_count);    // Liste
     if (mem)
       {
       assert((pos_list[0]>0) && (pos_list[0]<=BookmarkCount));
       DestMod=BookmarkDestName[pos_list[0]-1];		// hole Zielpunkt
       DestPos=BookmarkDestPos[pos_list[0]-1];
       BeamToPosAtMod(DestPos,DestMod);			// Gehe zu Zielpunkt
       XtFree((char*)pos_list);
       }
     break;
   case BMADD:		// Fuege neues Lesezeichen hinzu
     value_str = XmStringCreateSimple("");	// Prompt Dialog leeren
     ac=0;
     XtSetArg(al[ac],XmNtextString,value_str); ac++;
     XtSetValues(PromptDialog,al,ac);
     XmStringFree(value_str);
     XtManageChild(PromptDialog);		// Prompt Dialog oeffenen
     break;
   case BMDELETE:	// Loesche Lesezeichen
     assert(BookmarkCount>0);			// Hole Nr. und loesche Text-
     text=XmSelectionBoxGetChild(bookmarkbox,XmDIALOG_TEXT);	// Feld
     XmTextSetString(text,"");
     list=XmSelectionBoxGetChild(bookmarkbox,XmDIALOG_LIST);
     mem=XmListGetSelectedPos(list,&pos_list,&pos_count);
     if (mem)
       {
       assert((pos_list[0]>0) && (pos_list[0]<=BookmarkCount));
       BookmarkCount--;		// loesche intern
       if (BookmarkCount==0)
         {
         delete BookmarkDestPos;
         delete BookmarkDestName;
         }
       else
         {
         for (i=pos_list[0]-1; i<BookmarkCount; i++)
           {
           BookmarkDestPos[i]  = BookmarkDestPos[i+1];
           BookmarkDestName[i] = BookmarkDestName[i+1];
           }
         }
       XmListDeletePos(XmSelectionBoxGetChild(bookmarkbox,XmDIALOG_LIST),
                       pos_list[0]);	// loesche in Anzeige
       XtFree((char*)pos_list);
       }
     if (BookmarkCount==0)
       {
       XtSetSensitive(GotoButton,FALSE);
       XtSetSensitive(DeleteButton,FALSE);
       }
     break;
   case BMCANCEL:		// schliesse Lesezeichenfenster
     XtUnmanageChild(bookmarkbox);
     break;
   }
 }

static void PromptCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Eingabe in Prompt Dialog (Name fuer neues Lesezeichen)
 {
 int button = (int) client_data;
 XmSelectionBoxCallbackStruct* cbs = (XmSelectionBoxCallbackStruct*) call_data;
 char *s;
 long *p;
 char **q;
 long pos;
 int  len,i;

 switch (button)
   {
   case PDOK:		// neues Lesezeichen aufnehmen
     if (lastlink)
       {
       XmStringGetLtoR(cbs->value,XmSTRING_DEFAULT_CHARSET,&s); // Name
       len = strlen(s);
       XtFree(s);
       if (len==0) return;	// Kein Name eingegeben
       if (BookmarkCount == 0)	// Intern eintragen
         {
         BookmarkCount=1;
         BookmarkDestPos  = new long[BookmarkCount];
         BookmarkDestName = new char* [BookmarkCount];
         }
       else
         {
         p=BookmarkDestPos;
         q=BookmarkDestName;
         BookmarkCount++;
         BookmarkDestPos  = new long[BookmarkCount];
         BookmarkDestName = new char* [BookmarkCount];
         for (i=0; i<BookmarkCount-1; i++)
           {
           BookmarkDestPos[i]=p[i];
           BookmarkDestName[i]=q[i];
           }
         delete p;
         delete q;
         }
       lastlink->GetSource(&pos,&len);	// Mit aktuellem Symbol verbinden
       s = new char[strlen(lastmod)+1];
       strcpy(s,lastmod);
       BookmarkDestPos[BookmarkCount-1] = pos;
       BookmarkDestName[BookmarkCount-1]= s;
       XmListAddItem(XmSelectionBoxGetChild(bookmarkbox,XmDIALOG_LIST),
                     cbs->value,0);	// In Liste anzeigen
       if (BookmarkCount>0)
         {
         XtSetSensitive(GotoButton,True);
         XtSetSensitive(DeleteButton,True);
         }
       }
     break;
   case PDCANCEL:	// Kein neues Lesezeichen
     break;
   default: assert(0);
   };
 XtUnmanageChild(w);	// Prompt Dialog verstecken
 }

// Navigations-Window erzeugen
//----------------------------

void CreateNavWin(Widget toplevel)
 {
 Arg  al[15];
 int  ac;
 Widget form,sep;
 XmString label_str;

 // neue Shell fuer neues Window

 ac = 0;
 navshell = XtAppCreateShell("tTLA+ Browser Navigation","navwinshell",
                          applicationShellWidgetClass,
                          XtDisplay(toplevel),al,ac);

 // shell->form(1/1)   (i/j) = i-ter von j Soehnen

 ac=0;
 XtSetArg(al[ac],XmNwidth,400); ac++;
 XtSetArg(al[ac],XmNheight,100); ac++;
 form = XmCreateForm(navshell,"form",al,ac);
 XtManageChild(form);

 // shell->form->label(1/6)

 label_str = XmStringCreateSimple(" ");
 ac=0;
 XtSetArg(al[ac],XmNlabelString,label_str); ac++;
 XtSetArg(al[ac],XmNtopAttachment,XmATTACH_FORM); ac++;
 XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++;
 XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;
 XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNbottomPosition,50); ac++;
 label = XmCreateLabel(form,"label",al,ac);
 XtManageChild(label);
 XmStringFree(label_str);

 // shell->form->separator(2/6)

 ac=0;
 XtSetArg(al[ac],XmNtopAttachment,XmATTACH_WIDGET); ac++;
 XtSetArg(al[ac],XmNtopWidget,label); ac++;
 XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++;
 XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;
 XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_NONE); ac++;
 sep = XmCreateSeparator(form,"sep",al,ac);
 XtManageChild(sep);

 // shell->form->button(3/6)

 label_str = XmStringCreateSimple("Definition");
 ac=0;
 XtSetArg(al[ac],XmNlabelString,label_str); ac++;
 XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNtopPosition,55); ac++;
 XtSetArg(al[ac],XmNleftAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNleftPosition,2); ac++;
 XtSetArg(al[ac],XmNrightAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNrightPosition,23); ac++;
 XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNbottomPosition,95); ac++;
 button1 = XmCreatePushButton(form,"button1",al,ac);
 XtManageChild(button1);
 XtAddCallback(button1,XmNactivateCallback,NavigateCB,XtPointer(NAVDEF));
 XtSetSensitive(button1,False);
 XmStringFree(label_str);

 // shell->form->button(4/6)

 label_str = XmStringCreateSimple("Prev Use");
 ac=0;
 XtSetArg(al[ac],XmNlabelString,label_str); ac++;
 XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNtopPosition,55); ac++;
 XtSetArg(al[ac],XmNleftAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNleftPosition,27); ac++;
 XtSetArg(al[ac],XmNrightAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNrightPosition,48); ac++;
 XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNbottomPosition,95); ac++;
 button2 = XmCreatePushButton(form,"button2",al,ac);
 XtManageChild(button2);
 XtAddCallback(button2,XmNactivateCallback,NavigateCB,XtPointer(NAVPREV));
 XtSetSensitive(button2,False);
 XmStringFree(label_str);

 // shell->form->button(5/6)

 label_str = XmStringCreateSimple("Next Use");
 ac=0;
 XtSetArg(al[ac],XmNlabelString,label_str); ac++;
 XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNtopPosition,55); ac++;
 XtSetArg(al[ac],XmNleftAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNleftPosition,52); ac++;
 XtSetArg(al[ac],XmNrightAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNrightPosition,73); ac++;
 XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNbottomPosition,95); ac++;
 button3 = XmCreatePushButton(form,"button3",al,ac);
 XtManageChild(button3);
 XtAddCallback(button3,XmNactivateCallback,NavigateCB,XtPointer(NAVNEXT));
 XtSetSensitive(button3,False);
 XmStringFree(label_str);

 // shell->form->button(6/6)

 label_str = XmStringCreateSimple("Module");
 ac=0;
 XtSetArg(al[ac],XmNlabelString,label_str); ac++;
 XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNtopPosition,55); ac++;
 XtSetArg(al[ac],XmNleftAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNleftPosition,77); ac++;
 XtSetArg(al[ac],XmNrightAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNrightPosition,98); ac++;
 XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_POSITION); ac++;
 XtSetArg(al[ac],XmNbottomPosition,95); ac++;
 button4 = XmCreatePushButton(form,"button4",al,ac);
 XtManageChild(button4);
 XtAddCallback(button4,XmNactivateCallback,NavigateCB,XtPointer(NAVMOD));
 XtSetSensitive(button4,False);
 XmStringFree(label_str);

 XtRealizeWidget(navshell);
 }

// Bookmark-Window erzeugen
//-------------------------

int CreateBookmarkWin(Widget toplevel, char* specname)
 {
 Arg  al[10];
 int  ac;
 Widget button,list;
 XmString label_str,title_str,list_label_str,name_str;
 int  i,ok;
 char *name;

 // Bookmarkwindow ist ein SelectionDialog

 label_str = XmStringCreateSimple("Select a bookmark.");
 list_label_str = XmStringCreateSimple("Bookmarks");
 title_str = XmStringCreateSimple("tTLA+ Bookmarks");
 ac=0;
 XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
 XtSetArg(al[ac],XmNmustMatch,True); ac++;
 XtSetArg(al[ac],XmNselectionLabelString,label_str); ac++;
 XtSetArg(al[ac],XmNlistLabelString,list_label_str); ac++;
 XtSetArg(al[ac],XmNdialogTitle,title_str); ac++;
 bookmarkbox = XmCreateSelectionDialog(toplevel,"bookmarkbox",al,ac);
 XtAddCallback(bookmarkbox,XmNokCallback,BookmarkCB,XtPointer(BMGOTO));
 XtAddCallback(bookmarkbox,XmNapplyCallback,BookmarkCB,XtPointer(BMADD));
 XtAddCallback(bookmarkbox,XmNcancelCallback,BookmarkCB,XtPointer(BMDELETE));
 XmStringFree(label_str);
 XmStringFree(list_label_str);
 XmStringFree(title_str);

 // Goto Button

 GotoButton = XmSelectionBoxGetChild(bookmarkbox,XmDIALOG_OK_BUTTON);
 label_str = XmStringCreateSimple("Goto");
 ac=0;
 XtSetArg(al[ac],XmNlabelString,label_str); ac++;
 XtSetValues(GotoButton,al,ac);
 XmStringFree(label_str);

 // Add Button

 AddButton = XmSelectionBoxGetChild(bookmarkbox,XmDIALOG_APPLY_BUTTON);
 label_str = XmStringCreateSimple("Add");
 ac=0;
 XtSetArg(al[ac],XmNlabelString,label_str); ac++;
 XtSetValues(AddButton,al,ac);
 XmStringFree(label_str);
 XtSetSensitive(AddButton,False);

 // Delete Button

 DeleteButton = XmSelectionBoxGetChild(bookmarkbox,XmDIALOG_CANCEL_BUTTON);
 label_str = XmStringCreateSimple("Delete");
 ac=0;
 XtSetArg(al[ac],XmNlabelString,label_str); ac++;
 XtSetValues(DeleteButton,al,ac);
 XmStringFree(label_str);

 // Cancel Button

 button = XmSelectionBoxGetChild(bookmarkbox,XmDIALOG_HELP_BUTTON);
 label_str = XmStringCreateSimple("Cancel");
 ac=0;
 XtSetArg(al[ac],XmNlabelString,label_str); ac++;
 XtSetValues(button,al,ac);
 XmStringFree(label_str);
 XtAddCallback(button,XmNactivateCallback,BookmarkCB,XtPointer(BMCANCEL));

 // Prompt Dialog

 title_str = XmStringCreateSimple("Bookmark name.");
 label_str = XmStringCreateSimple("Enter bookmark name.");
 ac=0;
 XtSetArg(al[ac],XmNselectionLabelString,label_str); ac++;
 XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
 XtSetArg(al[ac],XmNdialogTitle,title_str); ac++;
 PromptDialog = XmCreatePromptDialog(toplevel,"prompt",al,ac);
 XtAddCallback(PromptDialog,XmNokCallback,PromptCB,XtPointer(PDOK));
 XtAddCallback(PromptDialog,XmNcancelCallback,PromptCB,XtPointer(PDCANCEL));
 XtUnmanageChild(XmSelectionBoxGetChild(PromptDialog,XmDIALOG_HELP_BUTTON));
 XmStringFree(label_str);
 XmStringFree(title_str);

 BookmarkCount=0;
 XtSetSensitive(GotoButton,False);
 XtSetSensitive(DeleteButton,False);

 if (! OpenBookmarkFile(specname,"r")) // Lesezeichen einlesen
   {
   return False;
   }
 BookmarkCount = GetNrOfBookmarks();
 if (BookmarkCount<0)
   {
   BookmarkCount=0;
   CloseBookmarkFile();
   return False;
   }
 if (BookmarkCount==0)
   {
   CloseBookmarkFile();
   return True;
   }
 BookmarkDestPos  = new long[BookmarkCount];
 BookmarkDestName = new char*[BookmarkCount];
 list=XmSelectionBoxGetChild(bookmarkbox,XmDIALOG_LIST);
 for (i=0; i<BookmarkCount; i++)
   {
   ok = ReadBookmark(&name,&(BookmarkDestName[i]),&(BookmarkDestPos[i]));
   if (!ok)
     {
     BookmarkCount=0;
     delete BookmarkDestPos;
     delete BookmarkDestName;
     CloseBookmarkFile();
     return False;
     }
   name_str = XmStringCreateSimple(name);
   delete name;
   XmListAddItem(list,name_str,0);	// zeilenweise anzeigen
   XmStringFree(name_str);
   }
 CloseBookmarkFile();
 XtSetSensitive(GotoButton,BookmarkCount>0);
 XtSetSensitive(DeleteButton,BookmarkCount>0);
 return True;
 }

// Bookmark-Window anzeigen
//-------------------------

void DisplayBookmarkWin()
 {
 XtManageChild(bookmarkbox);
 }

// NavigationsWindow loeschen
//---------------------------

void DestroyNavWin()
 {
 lastlink = NULL;
 if (navshell!=NULL) XtDestroyWidget(navshell);
 navshell = NULL;
 }

// BookmarkWindow loeschen
//---------------------------

void DestroyBookmarkWin()
 {
 int i;

 if (bookmarkbox!=NULL)
   {
   XtDestroyWidget(bookmarkbox);
   XtDestroyWidget(PromptDialog);
   }
 bookmarkbox = NULL;
 if (BookmarkCount>0)	// auch interne Struktur
   {
   for (i=0; i<BookmarkCount; i++)
     {
     delete BookmarkDestName[i];
     }
   delete BookmarkDestName;
   delete BookmarkDestPos;
   BookmarkCount=0;
   }
 }

// Lesezeichen in Datei sichern
//-----------------------------

int SaveBookmarks(char* specname)
 {
 Arg  al[10];
 int  ac;
 Widget list;
 XmStringTable itemTable;
 char* name;
 int i;

 if (! OpenBookmarkFile(specname,"w"))
   {
   return False;
   }
 if (BookmarkCount>0)
   {
   list=XmSelectionBoxGetChild(bookmarkbox,XmDIALOG_LIST);
   ac=0;
   XtSetArg(al[ac],XmNitems,&itemTable); ac++;
   XtGetValues(list,al,ac);
   for (i=0; i<BookmarkCount; i++)
     {
     XmStringGetLtoR(itemTable[i],XmSTRING_DEFAULT_CHARSET,&name);
     WriteBookmark(name,BookmarkDestName[i],BookmarkDestPos[i]);
     XtFree(name);
     }
   }
 CloseBookmarkFile();
 return True;
 }

// Anzeige in Navigations Window loeschen
//---------------------------------------

void ClearNavWin()
 {
 Arg al[10];
 int ac;
 XmString label_str;

 assert(navshell);
 lastlink = NULL;
 label_str = XmStringCreateSimple(" ");
 ac=0;
 XtSetArg(al[ac],XmNlabelString,label_str); ac++;
 XtSetValues(label,al,ac);
 XmStringFree(label_str);
 XtSetSensitive(button1,False);
 XtSetSensitive(button2,False);
 XtSetSensitive(button3,False);
 XtSetSensitive(button4,False);
 XtSetSensitive(AddButton,False);
 }

// Link im Navigations Window anzeigen
//------------------------------------

void DisplayLink(Link* l, char* SourceMod)
 {
 Arg al[10];
 int ac;
 XmString label_str;
 char *Mod;
 long pos;

 assert(navshell);
 XRaiseWindow(XtDisplay(navshell),XtWindow(navshell));
 lastlink=l;
 lastmod=SourceMod;
 label_str = XmStringCreateSimple(l->GetLinkText());
 ac=0;
 XtSetArg(al[ac],XmNlabelString,label_str); ac++;
 XtSetValues(label,al,ac);
 XmStringFree(label_str);
 l->GetDef(&pos,&Mod);
 XtSetSensitive(button1,pos!=-1);
 l->GetPrevUse(&pos,&Mod);
 XtSetSensitive(button2,pos!=-1);
 l->GetNextUse(&pos,&Mod);
 XtSetSensitive(button3,pos!=-1);
 l->GetMod(&pos,&Mod);
 XtSetSensitive(button4,pos!=-1);
 XtSetSensitive(AddButton,True);
 }
