/*----------------------------------------------------------------------*
 * Semconhelp.C
 *
 * Olaf Meier, 29.08.94
 *
 * Definiert die Hilfsklassen NumElem, NameElem, ActionVarsElem,
 * ProcessElem, Var_ProcessesElem, InterfaceElem fuer die semantische Konzepterfassung,
 * sowie einige Hilfsroutinen zur Konzepterstellung.
 *
 *----------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include "semconhelp.h"



NameElem::NameElem(char* n)  // der Name wird kopiert
    {
       name = new char[strlen(n) + 1];
       strcpy(name,n);
    }


ProcessElem::ProcessElem(char* n, List* pvl)
  {
     assert(n && pvl );

     processname = new char[strlen(n) + 1]; // Name wird kopiert.
     strcpy(processname,n);
     processvars = pvl;  // Achtung : Liste wird uebernommen!

  }

InterfaceElem::InterfaceElem(List* pl, List* svl)
  {
      assert(pl && svl);
      processes = pl;  // Achtung : Liste wird uebernommen!
      sharedvars = svl; // Achtung : Liste wird uebernommen!
  }


void AppendNumElem(int n, List* l)
 {
   assert(l);
   NumElem* numel = (NumElem*) l->GetFirstElem();
   while(numel)
    {
       if(numel->Getnumber() == n)    // Nummer schon vorhanden -> Abbruch
        return;
       numel = (NumElem*) l->GetNextElem();
    }
   numel = new NumElem(n);    // Nummer hinzufuegen
   l->Append(numel);
 }


void ConcatenateNumLists(List* l1,List* l2)
 { 
  assert(l1 && l2);
  NumElem* nel = (NumElem*) l1->GetFirstElem();
  while(nel)
   {
     AppendNumElem(nel->Getnumber(), l2);
     nel = (NumElem*) l1->GetNextElem();
   }
 }

void AppendNameElem(char* n, List* l)
 {
   assert(n && l);
   NameElem* nameel = (NameElem*) l->GetFirstElem();
   while(nameel)
    {
      if(strcmp(nameel->Getname() , n) == 0)    // Name ist schon vorhanden -> Abbruch!
       return;
      nameel = (NameElem*) l->GetNextElem();
    }
   nameel = new NameElem(n);  // Neuen Namen eintragen
   l->Append(nameel);
 }

void AppendActionVarsElem(int act, int var, List* l)
 { assert((act > 0) && (var > 0) && l);
   List* vl;
   ActionVarsElem* ae = (ActionVarsElem*) l->GetFirstElem();
   while(ae)
    {
       if(ae->GetactionID() == act)  // Aktion existiert bereits
        {
          vl = ae->Getvariables();  // Variablen zufuegen
          AppendNumElem(var, vl);
          return;
        }
       ae = (ActionVarsElem*) l->GetNextElem();
    }

   ae = new ActionVarsElem(act);   // Neue Aktion mit ihren Variablen hinzufuegen
   vl = ae->Getvariables();
   AppendNumElem(var, vl);
   l->Append(ae);
 }

void AppendVar_ProcessesElem(int var, char* proc,List* l)
 {
   assert((var > 0) && proc && l);
   List* pl;
   Var_ProcessesElem* vpe = (Var_ProcessesElem*) l->GetFirstElem();
   while(vpe)
    {
     if(vpe->GetvarID() == var)   // Variable existiert bereits
      {
        pl = vpe->Getprocesses();
        AppendNameElem(proc, pl);  // Prozesskomponenten zufuegen
        return;
      }
     vpe = (Var_ProcessesElem*) l->GetNextElem();
    }
   vpe = new Var_ProcessesElem(var);  // Neue Variable und
                                      // Prozesskomponenten  hinzufuegen
   pl = vpe->Getprocesses();
   AppendNameElem(proc, pl);      
   l->Append(vpe);
 }


void AppendInterfaceElem(List* pl, List* svl, List* l)
 { assert(pl && svl && l);
   List* variables;
   InterfaceElem* ife = (InterfaceElem*) l->GetFirstElem();
   while(ife)
    {
      variables = ife->Getsharedvars();
      if( SameSharedVariables(variables, svl))    // Interface existiert bereits
                                                  // mit diesen Variablen
       {
         delete pl;
         delete svl;
         return;
       }
      ife = (InterfaceElem*) l->GetNextElem();
    }
   ife = new InterfaceElem(pl, svl);   // Neue Interfacekomponente hinzufuegen
   l->Append(ife);
  }

int SameSharedVariables(List* l1, List* l2)
 {
   assert(l1 && l2);
   int found;
   NumElem *ne1 , *ne2;
   if(l1->GetLength() == l2->GetLength())
     {
       ne1 = (NumElem*) l1->GetFirstElem();
       while(ne1)
        {
          found = 0;
          ne2 = (NumElem*) l2->GetFirstElem(); 
          while(ne2)
           {
             if(ne1->Getnumber() == ne2->Getnumber())
              found = 1;
             ne2 = (NumElem*) l2->GetNextElem();
           }
          if(!found)
           return 0;  
          else 
           ne1 = (NumElem*) l1->GetNextElem();
         }
        return 1;
      }
    else
      return 0;
  }


int IsProcessSubset(List* l1, List* l2)
 { 
   assert(l1 && l2);
   NameElem *ne1, *ne2;
   int found;
   if(l1->GetLength() > l2->GetLength())
     return 0;
   else
    {
       ne1 = (NameElem*) l1->GetFirstElem();
       while(ne1)
        {
          found = 0;
          ne2 = (NameElem*) l2->GetFirstElem();
          while(ne2)
          {
            if(strcmp(ne1->Getname(), ne2->Getname()) == 0)
             found = 1;

            ne2 = (NameElem*) l2->GetNextElem();
          }
          if(!found)
           return 0;
          else
           ne1 = (NameElem*) l1->GetNextElem();
         }
        return 1;
     }
 }


void CollectActionvars(Node* n, List* avl,List* pvl, List* svl)
 {
    assert(n && avl && pvl && svl);
    Node* m;
    ActionVarsElem* ael;
    NumElem* nel;
    int actnr, varnr;
    List* variables;

    AppendNumElem(n->GetLevel(),svl);
    actnr = n->GetFirstActionNr();
    while(actnr != -1)
     {
       if(actnr > 0)
        {
          ael = (ActionVarsElem*) avl->GetFirstElem();
          while(ael)
           {
             if(ael->GetactionID() == actnr)   //  Aktion gefunden
              {
                variables = ael->Getvariables();  
                nel = (NumElem*) variables->GetFirstElem();
                while(nel)
                 { // Alle Variablen dieser Aktion werden der Liste 'pvl' zugefuegt
                    AppendNumElem(nel->Getnumber(), pvl);  
                    nel = (NumElem*) variables->GetNextElem();
                 }
                break; 
               }
              ael = (ActionVarsElem*) avl->GetNextElem();
            }
          }                  
       actnr = n->GetNextActionNr();
       }      
     m = n->GetFirstEdge();
     while(m)
      {
        CollectActionvars(m, avl, pvl, svl);
        m = n->GetNextEdge();
      }
  } 

void CollectSelectedVars(List* pvl, List* svl, AdjList* tree)
 { assert(pvl && svl && tree);
   Node *n;
   NumElem* pvar = (NumElem*) pvl->GetFirstElem();
   while(pvar)
    {
     n = tree->PTSearchforNodeLevel(pvar->Getnumber());
     while(n && (n->GetNodeType() == 'A' ) && n->HasPTpredecessor())
      { // ARRAY-Variable komplett mit saemtlichen Variablenkomponeten uebernehmen
        n = n->GetPTpredecessor();
      }
     SCTraverseTree(n,svl);
     pvar = (NumElem*) pvl->GetNextElem();
    }
 }

char* GetNameofNodeNr(int nr,AdjList* tree)
 {
   assert((nr > 0) && tree);
   Node* n = tree->PTSearchforNodeLevel(nr);
   if(n)
    return n->GetName();
   else
    return NULL;
 }  
 
void SCTraverseTree(Node* n, List* svl)
 {
    assert(n && svl);
    Node* m;
    AppendNumElem(n->GetLevel(), svl);
    m = n->GetFirstEdge();
    while(m)
     {
       SCTraverseTree(m, svl);
       m = n->GetNextEdge();
     }
  }

void CollectVar_ProcessesElems(List* pl, List* vpl)
 {
   assert(pl && vpl);
   List* varlist;
   NumElem* varnr;
   ProcessElem* pel = (ProcessElem*) pl->GetFirstElem();
   while(pel)
    {
     varlist = pel->Getprocessvars();
     varnr = (NumElem*) varlist->GetFirstElem();
     while(varnr)
      {
       AppendVar_ProcessesElem(varnr->Getnumber(), pel->Getprocessname(), vpl);
       varnr = (NumElem*) varlist->GetNextElem();
      }
     pel = (ProcessElem*) pl->GetNextElem();
    }
 }


void SearchforInterfaces(List* vpl, List* ifl)
 {
   assert(vpl && ifl);
   Var_ProcessesElem **vparray, *vpel;
   int vpl_length, i1, i2;
   NameElem* procname;
   NumElem* varnum;
   List *proclist,*newproclist, *varlist,*newvarlist;
 

   vpl_length = vpl->GetLength();
   if(vpl_length == 0) return;
   vparray = new Var_ProcessesElem*[vpl_length];
   i1 = 0;
   vpel = (Var_ProcessesElem*) vpl->GetFirstElem();
   while(vpel)
    {
      vparray[i1] = vpel;
      i1++;
      vpel = (Var_ProcessesElem*) vpl->GetNextElem();
    }

   for(i1 = 0; i1 < vpl_length ; i1++)
    {
      if(vparray[i1]->Getprocesses()->GetLength() > 1)
       {
          proclist = vparray[i1]->Getprocesses();
          newproclist = new List;
          newvarlist = new List;
          AppendNumElem(vparray[i1]->GetvarID(), newvarlist);
          procname = (NameElem*) proclist->GetFirstElem();
          while(procname)
           {
             AppendNameElem(procname->Getname(), newproclist);
             procname = (NameElem*) proclist->GetNextElem(); 
           }
          for(i2 = 0; i2 < vpl_length; i2++)
           {
              if(i1 != i2) 
               {
                  proclist = vparray[i2]->Getprocesses();
                  if(IsProcessSubset(newproclist,proclist))
                   AppendNumElem(vparray[i2]->GetvarID(), newvarlist);
               }  
           }
          AppendInterfaceElem(newproclist, newvarlist, ifl);
         }   
       }
     delete [] vparray ;
  }


void UnMarkVariables(List* vl,AdjList* tree)
 {
    assert(vl && tree);
    Node* n;
    NumElem* numel = (NumElem*) vl->GetFirstElem();
    while(numel)
     {
        n = tree->PTSearchforNodeLevel(numel->Getnumber());
        assert(n);
        n->Deselect();
        numel = (NumElem*) vl->GetNextElem();
     }
 }

void MarkVariables(List* vl,AdjList* tree)
 { 
   assert(vl && tree);
   Node* n;
   NumElem* numel = (NumElem*) vl->GetFirstElem();
   while(numel)
     {
        n = tree->PTSearchforNodeLevel(numel->Getnumber());
        assert(n);
        n->Select();
        numel = (NumElem*) vl->GetNextElem();
     }
 }


void UnMarkAllVariables(AdjList* tree)
 {
   assert(tree);
   Node *n = tree->GetFirstNode();
   while(n)
     {
         if(n->GetLevel() !=0)
           n->Resetselected();
         n = tree->GetNextNode();
     }
 }


AdjList* CreateProcInterfAdjList(List* pl,List* ifl)
 { assert(pl && ifl);
   ProcessElem* procel;
   InterfaceElem* ifel;
   NumElem* numel;
   NameElem* nameel;
   Node *n ,*last = NULL ;
   AdjList* PIadjlist;
   char *procname,*buttonname,  ifname[15];
   int interfacenr;
   
   
   List *varlist ,*proclist;

   if(pl->GetLength() == 0) return NULL;

   PIadjlist = new AdjList;
   procel = (ProcessElem*) pl->GetFirstElem();
 // Prozesskomponenten der Adjazenzliste zufuegen
 // Dabei werden die Prozessknoten (wie bei einer Perlenkette) hintereinandergehaengt.
   while(procel)
    {  
      procname = procel->Getprocessname();
      buttonname = new char[strlen(procname) + strlen("process ''") + 1];
      sprintf(buttonname,"process '%s'",procname);
      varlist = procel->Getprocessvars();
      numel = (NumElem*) varlist->GetFirstElem();
      while(numel)
       {
           n = PIadjlist->SCAddNode(buttonname,numel->Getnumber()); 
           numel = (NumElem*) varlist->GetNextElem();
        }
// Prozess-Kennung
      n->SetNodeType('P');
      if(last)
      last->AddEdge(n,'P');
      last = n;
      
      delete buttonname;
      procel = (ProcessElem*) pl->GetNextElem();
    }

    if(ifl->GetLength() == 0) return PIadjlist;
    interfacenr = 1;
// Interfacekomponente der Adjazenzliste zufuegen
    ifel = (InterfaceElem*) ifl->GetFirstElem();
    while(ifel)
     { 
        proclist = ifel->Getprocesses();
        varlist = ifel->Getsharedvars();
        numel = (NumElem*) varlist->GetFirstElem();
        sprintf(ifname,"interface %d",interfacenr);
        while(numel)
         {
           n = PIadjlist->SCAddNode(ifname, numel->Getnumber());
           numel = (NumElem*) varlist->GetNextElem();
         }
        n->SetNodeType('I');  // Interface-Kennung
        nameel = (NameElem*) proclist->GetFirstElem();
        while(nameel)
         { 
            procname = nameel->Getname();
            buttonname = new char[strlen(procname) + strlen("process ''") + 1];
            sprintf(buttonname,"process '%s'",procname);
            last = PIadjlist->GetFirstNode();
            while(last)
             {
               if(strcmp(last->GetName() , buttonname) == 0)
                last->AddEdge( n, 'I');
               last = PIadjlist->GetNextNode();
             }
            delete buttonname; 
            nameel = (NameElem*) proclist->GetNextElem();
         }
         
       interfacenr++;
       ifel = (InterfaceElem*) ifl->GetNextElem();
     }


    return PIadjlist;  
 }    
