/*----------------------------------------------------------------------*
 * Semconhelp.h
 *
 * Olaf Meier, 29.08.94
 *
 * Definiert die Hilfsklassen NumElem, NameElem, ActionVarsElem,
 * ProcessElem, Var_ProcessesElem, InterfaceElem fuer die semantische Konzepterfassung, 
 * sowie einige Hilfsroutinen zur Konzepterstellung. 
 * 
 *----------------------------------------------------------------------*/

#ifndef SEMCONHELP_H
#define SEMCONHELP_H

#include "list.h"
#include "adjlistcxx.h"



// Zahlen-Element einer Liste

class NumElem : public ListElem
{
  public :

   NumElem(int n)
    : number(n)
   {}

   ~NumElem()
    {}

   int Getnumber()
    {  return number;}

  private :

   int number;
};

// Name-Element einer Liste

class NameElem : public ListElem
 {
   public :

    NameElem(char* n);

    ~NameElem()
     { delete name;}

    char* Getname()
     { return name;}

   private :

    char* name;
 };

// Listenelement, das jeder Aktionsnummer die in dieser Aktion vorkommenden Variablennummern
// zuordnet

class ActionVarsElem : public ListElem
 {
   public :

    ActionVarsElem(int n)
     : actionID(n)
     {
       variables = new List;
     }

    ~ActionVarsElem()
     { delete variables;}

// Rueckgabe der (eindeutigen) Aktionsnummer
    int GetactionID()
     { return actionID; }

// Rueckgabe der Aktionsvariablen
    List* Getvariables()
     {
       return variables;
     }

   private :

      int actionID;
      List* variables;

 };


// Prozesselement zur Speicherung des gewaehlten Prozessnamens, sowie der fuer diesen
// Prozessnamen ausgewaehlten Variablen

class ProcessElem : public ListElem
 {
   public :

     ProcessElem(char* n, List* pvl);

     ~ProcessElem()
      {
        delete processname;
        delete processvars;
    
      }

// Rueckgabe des Prozessnamens

     char* Getprocessname()
       { return processname;}

// Rueckgabe der Prozessvariablen

     List* Getprocessvars()
       { return processvars;}



    private :

       char* processname;

       List* processvars;


 };


// Listenelement zur Speicherung einer Variablen und den Prozessnamen, die diese Variable
// nutzen. (dient zur Ermittlung geteilter Prozessvariablen) 

class Var_ProcessesElem : public ListElem
 {

  public :

     Var_ProcessesElem(int n)
      : varID(n)
     {
        processes = new List;
     }


     ~Var_ProcessesElem()
      {
        delete processes;
      }

// Rueckgabe der (eindeutigen) Variablennummer
     int GetvarID()
      { return varID;}

// Rueckgabe der betroffenen Prozessnamen   
     List* Getprocesses()
      { return processes;}

   private :

      int varID;
      List* processes;

 };

// Interfaceelement dient der Speicherung der geteilten Variablen, sowie den dazugehoerigen 
// Prozessnamen

class InterfaceElem : public ListElem
 {
   public :

      InterfaceElem(List* pl, List* svl);

      ~InterfaceElem()
       {
         delete processes;
         delete sharedvars;
       }

// Liefert Prozessnamensliste zurueck
      List* Getprocesses()
        {
          return processes;
	}

// Liefert Liste geteilter Variablen zurueck

      List* Getsharedvars()
        {
           return sharedvars;
	}

      private :

        List* processes;
        List* sharedvars;
 };


// Fuegt Zahl 'n' zur Zahlenliste hinzu, falls diese nicht in 'l' enthalten ist.
void AppendNumElem(int n, List* l);

// Verschmelzt zwei Zahlenliste 'l1','l2'
void ConcatenateNumLists(List* l1,List* l2);

// Fuegt Namen 'n' zur Namensliste hinzu, falls dieser nicht in 'l' enthalten ist.
void AppendNameElem(char* n, List* l);

// Fuegt Aktionsnummer 'n' mit der Variablennummer 'var' an die Liste 'l', falls diese nicht 
// vorhanden sind. 
void AppendActionVarsElem(int act, int var, List* l);

// Fuegt Variablennummer 'var' mit dem Prozessnamen 'proc' an die Liste 'l', falls diese nicht
// vorhanden sind.// 
void AppendVar_ProcessesElem(int var, char* proc,List* l);

// Fuegt Namensliste 'pl' mit dazugehoeriger Variablenliste 'svl' zur Liste 'l' hinzu, falls
// Liste  'svl'  dort nicht vorhanden ist.
void AppendInterfaceElem(List* pl, List* svl, List* l);

// Ueberpruefung, ob Variablenlisten die gleichen Variablennummer enthalten (1 gleich, 0 sonst)
int SameSharedVariables(List* l1, List* l2);

// Ueberpruefung, ob Namensliste 'l1' Teilmenge der Liste 'l2' ist (1 Teilmenge, 0 sonst)
int IsProcessSubset(List* l1,List* l2);

// Sammelt die Aktionskennungen jeder Systemvariable im Variablenbaum  vom Knoten 'n' an.
// Die betreffenden Variablen der Aktionen werden in der Variablenliste 'svl' 
// sowie in Variablenliste 'pvl' eingetragen.
void CollectActionvars(Node* n, List* avl,List* pvl, List* svl);

// Jede Systemvariable der Liste 'pvl' wird im Variablenbaum 'tree' gesucht.
// Anschliessend werden die Variablen des betreffenden Teilbaums (der gefundenen
// Variable)  der Variablenliste
// 'svl' zugefuegt.
void CollectSelectedVars(List* pvl, List* svl, AdjList* tree);

// Gibt den Namen des jeweiligen Knotens mit der Knotennummer 'nr' im Baum 'tree' zurueck.
char* GetNameofNodeNr(int nr,AdjList* tree);

// Traversierung des vom Knoten 'n' gebildeten Teilbaums. Alle besuchten Varablenknoten
// werden in die Liste 'svl' eingetragen.
void SCTraverseTree(Node* n, List* svl);

// Aus der Prozesskomponten-Liste wird eine neue Uebersicht erstellt, die jeder Variablen
// ihre zugeordneten Prozesskomponenten gegenueberstellt. Diese Uebersicht wird in 
// der Liste 'vpl' vermerkt. 
void CollectVar_ProcessesElems(List* pl, List* vpl);


// Mithilfe der Variablen->Prozeslisten Uebersicht ermittelt diese Routine
// die Liste 'ifl' aller vorhandenen Interfacekomponenten.
void SearchforInterfaces(List* vpl, List* ifl);


// Vermindert den 'selected' Zaehler im Knoten aller Variablen der Liste 'vl' innerhalb der 
// Adjazenzlisten-Darstellung 'tree' um 1.
void UnMarkVariables(List* vl,AdjList* tree);

// Erhoeht den 'selected' Zaehler im Knoten aller Variablen der Liste 'vl' innerhalb der
// Adjazenzlisten-Darstellung 'tree' um 1.
void MarkVariables(List* vl,AdjList* tree);

// Setzt den selected' Zaehler aller Variablenknoten in 'tree' auf 0 zurueck.
void UnMarkAllVariables(AdjList* tree);


// Diese Routine transformiert die Prozessliste 'pl' und die Interfaceliste 'ifl' 
// in eine gemeinsame Adjazenzlisten-Darstellung, die zurueckgeliefert wird.
AdjList* CreateProcInterfAdjList(List* pl,List* ifl);


#endif
