/* scanner.l ( urspruenglich tla+.l )
   lex-file for TLA+
   
   (E) 2.0,  9.7.92,  Arnulf Mester
   Alle Rechte gem. dt. UrhG und WissHG vorbehalten.
   Jedwede Verwendung nur mit schriftlicher Genehmigung des Autors.
*/

/*****************************************************************************
 * TLA+ Browser : Praeprozessor - Scanner
 *
 * Michael Geikowski, 30.6.93
 *
 * Olaf Meier, 01.09.94
 *
 * Folgende Aenderungen vorgenommen: 
 *
 * Routine intvalue() geandert (idvalue() statt token()-Routine)
 * Routine tupleselectorvalue() ergaenzt
 * Routine stringvalue() geandert (idvalue() statt token()-Routine)
 *
 *****************************************************************************/

%{
#include <string.h>
#include "symlist.h"
#include "dbinterface.h"
#include "parser2.tab.h"

#define LEXDEBUG   1
#define TABLENGTH  4
#define MAXIPS    80

extern int yylex();
extern char* yytext;
extern SymListElem* yypplval;

int lex_list  = 0;	// no (0) output of actuell line scanned
int debug_ips = 0; 	// no (0) debugging of indentation routines
int writeit   = 0;	// kein (0) Echo der token nach outfile
SymList* sList;         // Liste der gefundenen token
SymListElem* lastElem = 0; // letztes token

static int columnno = 1;	// aktuelle Spaltennr
static int pos = 0;		// aktuelle Position (in char seit Fileanfang)
static int noPrefixAO=0;	// status: ein prefAO vorbei auf dieser Zeile?
static int i;			// auxvar to remember yyleng in some rules
static int indentPosStack[MAXIPS];	// Stack
static int indentPosStackPtr = 0;	// Stackpointer
static char* actModName;	// Name des gescannten Moduls

static int ips_top()
 {
 if ( indentPosStackPtr == 0 ) return -1 ; 	/* return error code */
 else
   { 
   if(debug_ips) {
     fprintf(stderr,"ips_top called, returned %d from loc %d \n",
                     indentPosStack[indentPosStackPtr-1], indentPosStackPtr-1);
     fflush(stderr); }
   return indentPosStack[indentPosStackPtr-1] ;
   }
 }

static int ips_push( int newElement )
 {
 if(debug_ips) {
   fprintf(stderr, "ips_push( %d ) called\t", newElement);
   fflush(stderr); }
 if(columnno==ips_top()) {
   fprintf( stderr, "scanner warning: opening structures on same indentation in line %d\n", lineno);
   fflush(stderr); }
 if ( indentPosStackPtr > MAXIPS ) {
   fprintf(stderr,"ips stacklimit exceeded \n");
   fflush(stderr);
   exit(-1); }
 else { 
   indentPosStack[indentPosStackPtr] = newElement;
   if (debug_ips) {
     fprintf(stderr,"at loc %d \n", indentPosStackPtr);
     fflush(stderr); }
   indentPosStackPtr ++;
   return 0; }
 return 0;
 }
	  
static void ips_pop()
 {
 if ( indentPosStackPtr != 0 )/* only do something if stack is filled */
   { 
   indentPosStackPtr --;
   if(debug_ips) {
     fprintf(stderr,"ips_pop called, removed %d from loc %d \n",
                    indentPosStack[indentPosStackPtr], indentPosStackPtr);
     fflush(stderr); }
   }
 }
  	  
static void change_noPrefixAO(int tok)
      /* Nach diesen Token sind Prefix-AND, Prefix-OR moeglich */ 
 {
 switch (tok)
   {
   case IBL:
   case IBR:
   case PREF_AND:
   case PREF_OR:
   case REPLACE:
   case RBR_ON:
   case '[':
   case LT:
   case COMMA:
   case COLON:
   case SEMICOLON:
   case _LEADSTO_:
   case _L_AND_:
   case _L_OR_:
   case DEFINED_AS: noPrefixAO = 0; break;

   default:         noPrefixAO = 1; break;
   };
 }

static int token (int tok,int leng)
       /* Normales Token behandeln */
 {
 if (leng > 0)	/* ungleich IBL,IBR */
   {
   yylval.selem = new SymListElem (pos,leng,lineno,tok); // ->om: Zusaetzlicher Parameter
   yypplval = yylval.selem;
   sList->Append(yylval.selem);
   lastElem = yylval.selem;
   }
 else
   {
   yylval.selem = 0;	/* IBL,IBR */
   yypplval = yylval.selem;
   }
 pos      += leng;
 columnno += leng;
 change_noPrefixAO(tok);
 if (writeit && (tok!=DEFINED_AS) )	/* Inhaltsverzeichnis */
   {
   AddToContentsLine(yytext);
   }
 return tok;
 }

static int tabcol (int col)
      /* Tabulator behandeln */
{
 int hcol = (col - 1) / TABLENGTH;
 return ((hcol + 1) * TABLENGTH) + 1;
 }

static void comment_handle (char* string, int leng)
 {
 char hel = 0;
 for (int i = 0; i < leng; i++)
   {
   hel = string [i];
   if ((hel == '\n') || (hel == '\r'))
     {
     pos++;
     columnno = 1;
     lineno++;
     }
   else
     if (hel == '\t')
       {
       pos++;
       columnno = tabcol (columnno);
       }
     else
       {
       pos++;
       columnno++;
       }
   }
 }



static int realvalue (int tok, int leng)		/* unused */
 {
 return token(tok,leng);
 }

static char* ptr_strdup (char* s)
        /* Hilfsfunktion */
 {
 char* p = new(char[strlen(s)+1]);
 strcpy (p ,s);
 return p;
 }



static int idvalue (int tok, int leng)
       /* Token normalID behandeln */
 {
 char* hstr = ptr_strdup (yytext);
 char primed = hstr[strlen(hstr)-1] == '\'';
 if (primed)
   {
   hstr[strlen(hstr)-1] = '\0';
   }
 id_info* idp = new id_info(hstr,nidt_normal,primed);
 yylval.selem = new SymListElem (pos,strlen(hstr),lineno,tok,idp); // ->om: Zusaetzlicher Parameter
 yylval.selem->SetDefiningModule(actModName);
 yypplval = yylval.selem;
 sList->Append(yylval.selem);
 lastElem = yylval.selem;
 pos      += leng;
 columnno += leng;
 change_noPrefixAO(tok);
 if (writeit && (tok!=DEFINED_AS) )    /* Inhaltsverzeichnis */
   {
   AddToContentsLine(yytext);
   }
 return tok;
 }

static int stringvalue (int tok, int leng) /* ->om */
        /* Token STRING behandeln */
 {
 return idvalue(tok,leng);
 }

static int tupleselectorvalue(int tok, int leng) /* ->om */
        /* Token TUPLESELECTOR behandeln */
 {
 return idvalue(tok,leng);
 }

static int intvalue(int tok, int leng) /* ->om */
        /* Token NATNUM behandeln */
 {
 return idvalue(tok,leng);
 }

%}

D		[0-9]
E		[DEde][-+]?{D}+
nostarp1	([^*]|\*+[^*)])
nostarp2	([^*]|\*+[^*\/])
white		[ \t]
alpha		[a-zA-Z]
alphanum	[a-zA-Z0-9]
X		[Xx]
hex		[0-9a-fA-F]

%x doIBL search
%%

^#abort$			{ fprintf(stderr,"user abort\n");
                                  fflush(stderr);
                                  exit(-1);
                                }
^.*$				|
<search>^.*$			{ if(lex_list) {
                                    fprintf(stderr,"\n* %d: %s\n",lineno,
                                                                     yytext);
                                    fflush(stderr); }
				  yyless(0); REJECT;
                                }

"=="				return token(DEFINED_AS, yyleng);
"="				return token(EQ, yyleng); 
"#"				|
"!="				return token(NE, yyleng);
"<="				return token(LE, yyleng);
"<"				return token(LT, yyleng);
">="				return token(GE, yyleng);
">"				return token(GT, yyleng); 
"+"				return token(PLUS, yyleng); 
"-"				return token(MINUS, yyleng);
"*"				return token(TIMES, yyleng); 
"/"				return token(DIVIDED, yyleng);
"DIV"				return token(DIV, yyleng); 
"MOD"				return token(MOD, yyleng); 
"^"				return token(POWER, yyleng); 
":forall:"			return token(_ALL_QUANTOR_, yyleng); 
:[xX]:				return token(_CROSS_, yyleng);
":A:"				return token(_ALL_QUANTOR_, yyleng); 
":exists:"			return token(_EXISTS_QUANTOR_, yyleng);
":EE:"				return token(_BOLD_EXISTS_, yyleng); 
":E:"				return token(_EXISTS_QUANTOR_, yyleng);
\[\]				|
":henceforth:"			return token(_HENCEFORTH_, yyleng); 
"<>"				|
":eventually:"			return token(_EVENTUALLY_, yyleng);
"~>"				|
":leadsto:"			return token(_LEADSTO_, yyleng);

"'"				return token(PRIME, yyleng); 
":in:"				return token(_ELEMENT_OF_, yyleng); 
":not_in:"			return token(_NOT_ELEMENT_OF_, yyleng);
":cup:"				|
":union:"			return token(_UNION_, yyleng);	
":intersection:"		return token(_INTERSECT_, yyleng);
":subset:"			return token(_SUBSET_OF_, yyleng);
"{}"				|  
":emptyset:"                    return token(_EMPTYSET_, yyleng); 
"\\"				return token(_SET_DIFFERENCE_, yyleng);

WF				return token(WF, yyleng);
SF				return token(SF, yyleng);

[Ss]et				|
SET				return token(SET, yyleng); 

[Tt]rue				|
TRUE				return token(TRUE, yyleng); 
[Ff]alse			|
FALSE				return token(FALSE, yyleng); 

[Ii]nfix			|
INFIX				return token(INFIX, yyleng); 
[Ii]nfix[Rr]			|
INFIXR				return token(INFIXR, yyleng); 

[Ee]nabled			|
"ENABLED"			return token(ENABLED, yyleng); 
[Uu]nchanged			|
"UNCHANGED"			return token(UNCHANGED, yyleng);
[Uu]nion			|		
"UNION"				return token(UNION, yyleng); 
[Ss]ubset			|
"SUBSET"			return token(SUBSET, yyleng); 
[Cc]hoose			|
"CHOOSE"			return token(CHOOSE, yyleng); 
[Dd]omain			|
DOMAIN				return token(DOMAINt, yyleng);
"->"                            return token(_RIGHTARROW_, yyleng);
"|->"                           |
":mapsto:"			return token(_MAPSTO_, yyleng); 
"CHANGE"			|
"CHG"				return token(CHANGE, yyleng);

%{ /* 
    * tokens which open an invible brace 
    */
%}

"IF"				{ i=yyleng;
                                  yyless(0);
                                  BEGIN(doIBL);
                                  return token(IF,i);
                                } 
<doIBL>"IF"			{ if(columnno>=ips_top())
                                    { ips_push(columnno);
                                      BEGIN(INITIAL);
                                      return token(IBL,0); }
                                }
"THEN"				{ i=yyleng;
                                  yyless(0);
                                  BEGIN(doIBL);
                                  return token(THEN,i);
                                } 
<doIBL>"THEN"			{ if(columnno>=ips_top())
                                    { ips_push(columnno);
                                      BEGIN(INITIAL);
                                      return token(IBL,0); }
                                }
"ELSE"				{ i=yyleng;
                                  yyless(0);
                                  BEGIN(doIBL);
                                  return token(ELSE,i);
                                } 
<doIBL>"ELSE"			{ if(columnno>=ips_top())
                                    { ips_push(columnno);
                                      BEGIN(INITIAL);
                                      return token(IBL,0); }
                                }

"CASE"				{ i=yyleng;
                                  yyless(0);
                                  BEGIN(doIBL);
                                  return token(CASE,i);
                                } 
<doIBL>"CASE" 			{ if(columnno>=ips_top())
                                    { ips_push(columnno);
                                      BEGIN(INITIAL);
                                      return token(IBL,0);}
                                }

"LET"				{ i=yyleng;
                                  yyless(0);
                                  BEGIN(doIBL);
                                  return token(LET,i);
                                } 
<doIBL>"LET"			{ if(columnno>=ips_top())
                                    { ips_push(columnno);
                                      BEGIN(INITIAL);
                                      return token(IBL,0);}
                                }
"IN" 				{ i=yyleng;
                                  yyless(0);
                                  BEGIN(doIBL);
                                  return token(IN,i);
                                } 
<doIBL>"IN" 			{ if(columnno>=ips_top())
                                    { ips_push(columnno);
                                      BEGIN(INITIAL);
                                      return token(IBL,0); }
                                }

\~				|
":not:"				{ i=yyleng;
                                  yyless(0);
                                  BEGIN(doIBL);
                                  return token(_L_NOT_,i);
                                } 
<doIBL>\~			|
<doIBL>":not:"			{ if(columnno>=ips_top())
                                    { ips_push(columnno);
                                      BEGIN(INITIAL);
                                      return token(IBL,0);}
                                }

"/\\"				|
":and:"				{ if(noPrefixAO)
                                    { return token(_L_AND_,yyleng); }
				  else
                                    { i=yyleng;
                                      yyless(0);
                                      BEGIN(doIBL);
                                      /* noPrefixAO=1; */
                                      return token(PREF_AND,i);}
				} 
<doIBL>"/\\"			|
<doIBL>":and:"			{ if(columnno>=ips_top())
                                    { ips_push(columnno);
                                      BEGIN(INITIAL);
                                      return token(IBL,0);}
                                }

"\\/"				|
":or:"				{ if(noPrefixAO)
                                    { return token(_L_OR_, yyleng); }
				  else
                                    { i=yyleng;
                                      yyless(0);
                                      BEGIN(doIBL);
                                      /* noPrefixAO=0; */
                                      return token(PREF_OR,i); }
				}
<doIBL>"\\/"			|
<doIBL>":or:"			{ if(columnno>=ips_top())
                                    { ips_push(columnno);
                                      BEGIN(INITIAL);
                                      return token(IBL,0);}
                                }

"=>"				|
":implies:"			{ i=yyleng;
                                  yyless(0);
                                  BEGIN(doIBL);
                                  return token(_IMPLIES_,i);
                                } 
<doIBL>"=>"			|
<doIBL>":implies:"		{ if(columnno>=ips_top())
                                    { ips_push(columnno);
                                      BEGIN(INITIAL);
                                      return token(IBL,0);}
                                }

"<=>"                           |
":equiv:"			{ i=yyleng;
                                  yyless(0);
                                  BEGIN(doIBL);
                                  return token(_EQUIV_,i);
                                } 
<doIBL>"<=>"                    |
<doIBL>":equiv:"		{ if(columnno>=ips_top())
                                    { ips_push(columnno);
                                      BEGIN(INITIAL);
                                      return token(IBL,0);}
                                }

"::"				{ i=yyleng;
                                  yyless(0);
                                  BEGIN(doIBL);
                                  return token(DOUBLECOLON,i);
                                } 
<doIBL>"::"			{ if(columnno>=ips_top())
                                    { ips_push(columnno);
                                      BEGIN(INITIAL);
                                      return token(IBL,0);}
                                }
 
"MODULE"			return token(MODULE, yyleng); 
"END"				return token(END, yyleng); 
"IMPORT"			return token(IMPORT, yyleng); 
"AS"				return token(AS, yyleng); 
"EXPORT"			return token(EXPORT, yyleng); 
"OPAQUE"			return token(OPAQUE, yyleng); 
"PARAMETERS"			return token(PARAMETERS, yyleng);
"CONSTANTS"			return token(CONSTANTS, yyleng);
"CONSTANT"			return token(CONSTANT, yyleng);
"CONST"				return token(CONSTANT, yyleng);
"BOOLEANS"			return token(BOOLEANS, yyleng);
"BOOLEAN"			return token(BOOLEAN, yyleng); 
"BOOL"				return token(BOOLEAN, yyleng); 
"PREDICATES"			return token(PREDICATES, yyleng);
"PREDICATE"			return token(PREDICATE, yyleng); 
"STATE"				return token(STATE, yyleng); 
"TRANSITION"			return token(TRANSITION, yyleng);
"FUNCTIONS"			return token(FUNCTIONS, yyleng); 
"FUNCTION"			return token(FUNCTION, yyleng); 
"FCN"				return token(FUNCTION, yyleng); 
"ACTIONS"			return token(ACTIONS, yyleng); 
"ACTION"			return token(ACTION, yyleng); 
"TEMPORALS"			return token(TEMPORALS, yyleng); 
"TEMPORAL"			return token(TEMPORAL, yyleng);
"INCLUDE"			return token(INCLUDE, yyleng);
"WITH"				return token(WITH, yyleng); 
"<-"				return token(REPLACE, yyleng);
"ASSUMPTIONS"			return token(ASSUMPTIONS, yyleng); 
"THEOREMS"			return token(THEOREMS, yyleng);
"Nat"				return token(NAT, yyleng);
ANY				|
[Aa]ny				return token(ANY, yyleng);

"\."				return token(DOT, yyleng); 
","				return token(COMMA, yyleng);
":"				return token(COLON, yyleng); 
";"				return token(SEMICOLON, yyleng);
"("				return token(RBR_ON, yyleng); 
")"				return token(RBR_OFF, yyleng); 
\[\[				return token(DSBR_ON, yyleng); 
\]\]				return token(DSBR_OFF, yyleng); 

{alpha}{alphanum}*		return idvalue(normalID,yyleng);
{alpha}{alphanum}*'		return idvalue(normalID,yyleng);

\"[^\"]*\"			return stringvalue(STRING, yyleng); 

\(\*{nostarp1}*\*+\){white}*	{ comment_handle (yytext, yyleng);
				  /* specificator's comments */
				}
\/\*{nostarp2}*\*+\/{white}*	{ comment_handle (yytext, yyleng);
				  /* mechanically generated comments 
				     (e.g. emacs TLA+ major mode zooming fct)*/
				}

\.{D}+				return tupleselectorvalue(TUPELSELECTOR,yyleng);
                                  /* TODO: has to return the value! */
{D}+				return  intvalue(NATNUM,yyleng);
                                  
%{ 
   /* 0{X}{hex}+			return hexvalue ... */
%}  

"\n"				|
"\r"				{ noPrefixAO=0;
				  pos += yyleng;
				  columnno = 1;
				  lineno++;
				  BEGIN(search);
				}

<search>" "			{ columnno += yyleng;
				  pos += yyleng;
				}
<search>"\t"			{ columnno = tabcol (columnno);
				  pos++;
				}
<search>"\n"                	|
<search>"\r"			{ pos+=yyleng; columnno = 1; lineno++;
                                }
<search>[^ \t]			{ yyless(0);
                                  if(columnno<=ips_top())
                                    { ips_pop(); return token(IBR,0); }
                                  else
                                    { BEGIN(INITIAL); };
				  if(debug_ips) {
                                    fprintf(stderr,"search: col %d <= top %d ?\n", columnno, ips_top());
                                    fflush(stderr); }
				}

" "				{ columnno += yyleng;
				  pos += yyleng;
                                  if(debug_ips) {
                                   fprintf(stderr,"new col: %d --",columnno);
                                   fflush(stderr); }
                                }
"\t"				{ columnno = tabcol (columnno);
				  pos++;
				}

.          	                return token(yytext[0],1);

%%

void reinitialize_lex (FILE* ifp, char* filename)
 {
 lineno     = 1;
 columnno   = 1;
 pos        = 0;
 writeit    = 0;
 noPrefixAO = 0;
 lastElem   = 0;
 sList      = new SymList;
 indentPosStackPtr = 0;
 actModName = filename;

 yy_init    = 1;
 yyin       = ifp;
}

int yypplex()
     /* Alias fuer Pass-1-Parser */
 {
 return yylex();
 }
