/*-------------------------------------------------------------------------
 * Semconcept.C
 *
 * Olaf Meier,  10.9.94
 *
 * Dieses Modul bereitet die zur semantischen Konzepterfassung notwendigen Informationen
 * auf. U.a. ist es fuer die Expansion der Systemformel zustaendig. Des weiteren werden
 * die in der Systemformel auftretenden Aktionen ermittelt. Diese Aktionen mit den jeweils
 * genutzten Systemvariablen werden abschliessend in eine Systemvariablen-Datei zwischen-
 * gespeichert, damit sie spaeter von der Browserkomponente gelesen werden koennen.
 *
 *--------------------------------------------------------------------------------------------*/

#include <assert.h>
#include <string.h>

#include "semconcept.h"

#include "sort.h"
#include "crossref.h"
#include "parser2.tab.h"


// Externe Variable
extern CrossRef* crossref; 


// Modulglobale Variablen
ParseExpr *pe_semantic_concept = NULL, *pe_variables = NULL;
static ParseExpr* actualOpDef = NULL;
static int ActionNr = 0;
static SymList *symbols = NULL;


// Hilfsklasse zur Speicherung von Srings

class NameElem : public ListElem
 {
   public :
   NameElem(char* n)
    : name(n)
   {};
   ~NameElem()
   {};
   char* GetName()
    { return name; };

   private : 
    char *name;
  }; 
   
        

void GetSubstitutions(ParseExpr* expr,ParseExpr* start, List* substlist,SymList* sl)
 { assert(expr && start && substlist && sl);
   IDinfo* substinfo;
   if(expr->GetMaxIndex() == 0) // eine Substitution
    {
      substinfo = new IDinfo;
      substinfo->SetParname(expr->GetPE(1)->GetSE(1)->Getlexem());
      ExpandExpression(expr->GetPE(1)->GetPE(3),start,sl); 
      substinfo->SetExpression(expr->GetPE(1)->GetPE(3));
      substlist->Append(substinfo);
    }
   else // weitere Substitutionen 
    {
      GetSubstitutions(expr->GetPE(1),start,substlist,sl);
      substinfo = new IDinfo;
      substinfo->SetParname(expr->GetPE(3)->GetSE(1)->Getlexem());
      ExpandExpression(expr->GetPE(3)->GetPE(3),start,sl);
      substinfo->SetExpression(expr->GetPE(3)->GetPE(3));
      substlist->Append(substinfo);
    }
 }

List* CollectSubstitutions(ParseExpr* expr,ParseExpr* start,SymList* sl)
 {  assert(expr && start && sl);
    List* subst; 

    if(expr->GetMaxIndex() == 1) // Parametersubstitutionen sind vorhanden
      { 
         subst = new List;
         GetSubstitutions(expr->GetPE(2),start,subst,sl);
         return subst;  
      }
    else
      return NULL; 
 }       

void ParameterSubstitution(ParseExpr* expr, List* substlist,SymList* sl)
 { assert(expr && substlist && sl);
   int maxi = expr->GetMaxIndex();
   Symbol_Type st;
   noTerminal_Type nt;
   SymListElem* se; 
   ParseExpr *pe,*pecopy,*penew,*pesubst;;
   ParseSymb* ps;
   IDinfo *idel;
   for(int i = 1 ; i <= maxi+1; i++)
    {
      ps = expr->GetExpr(i);
      st = ps->GetSymbolType();
      switch(st)
       {
         case Empty : break;
         case Terminal : break;
         case NonTerm : pe = ps->GetParseExpr();
                   nt = pe->GetNonTermType();
                   if (nt == tID)                 // benutzerdefinierter Bezeichner
                     
                    {
                      se = pe->GetSE(2);
                      idel = (IDinfo*) substlist->GetFirstElem();
                      while(idel)
                       {
                         if(strcmp(idel->GetParname(), se->Getlexem()) == 0)  // ...Bezeichner taucht in
                                                                              // Substitutionsliste auf
                           {  
                              pesubst = idel->GetExpression();
                              if( (pesubst->GetMaxIndex() == 0) &&
                                  (pesubst->GetExpr(1)->GetSymbolType() == NonTerm) &&
                                  (pesubst->GetPE(1)->GetNonTermType() == tSetExpr) &&
                                  (pesubst->GetPE(1)->GetMaxIndex() == 0) &&
                                  (pesubst->GetPE(1)->GetExpr(1)->GetSymbolType() == NonTerm) &&
                                  (pesubst->GetPE(1)->GetPE(1)->GetNonTermType() == tID))
                                  
                               {

                                 pecopy = CopyTree(pesubst->GetPE(1)->GetPE(1),sl);
                                 delete pe;
                                 PopandAppend(pecopy,expr,i); // Bezeichner durch Bezeichner austauschen
                               }      
                              else
                               {
                                 penew = new ParseExpr(substituted,"substituted",1);
                                 pecopy = CopyTree(pesubst,sl);                             


                                 PopandAppend(pecopy,penew,1);  // Bezeichner durch Ausdruck austauschen

                                 delete pe;
                                 PopandAppend(penew,expr,i);  // und im Baum eingliedern

                               }
                            }
                         idel = (IDinfo*) substlist->GetNextElem();
                        }
                      }
                    else
                      ParameterSubstitution(pe, substlist,sl);
                    break;
           default : assert(0);
           }
        }
      }

                  
void TreeImport(ParseExpr* tree,ParseExpr* start,SymList* sl)
 { assert(tree && start && sl); 
   int maxi =  tree->GetMaxIndex();          
   noTerminal_Type nt;
   Symbol_Type st;
   ParseSymb* ps;
   ParseExpr *pe, *treecopy;
   SymListElem* se;
   List* substlist;
   
   for(int i = 1; i <= maxi+1 ; i++)
    {
      ps = tree->GetExpr(i);
      st = ps->GetSymbolType();
      switch(st)
       {
         case Empty : break;
         case Terminal : break;
         case NonTerm : pe = ps->GetParseExpr();
                        nt = pe->GetNonTermType();   
                        if(nt == tImportList)   // Importliste vorhanden
                          { 
                            if(pe->GetPE(1)->GetNonTermType() == tModuleName)   
                              {
                                 se = pe->GetPE(1)->GetSE(1);
                                 treecopy = crossref->GetParseTree(se->Getlexem());
                                 if(treecopy)
                                  {
                                   treecopy = CopyTree(treecopy,sl);
                                   pe->GetPE(1)->AddExpr(treecopy,1);  // Ableitungsbaum des includierten
                                                                       // Moduls einhaengen
                                  }
                              } 
                            else if(pe->GetPE(3)->GetNonTermType() == tModuleName)
                              {
                                 TreeImport(pe,start,sl);
                                 se = pe->GetPE(3)->GetSE(1);
                                 treecopy = crossref->GetParseTree(se->Getlexem());
                                 if(treecopy)
                                  {
                                    treecopy = CopyTree(treecopy,sl);
                                    pe->GetPE(3)->AddExpr(treecopy,1);
                                  }
                                
                              }
                           } 
                          else if(nt == tIncludePart)   // Include-Anweisung
                           { 
                                 // evtl. Parametersubstitutionen ermitteln
                                 substlist = CollectSubstitutions(tree->GetPE(i+1),start,sl);
                                  
                                 se = pe->GetPE(1)->GetSE(1);
                                 treecopy = crossref->GetParseTree(se->Getlexem());
                                 if(treecopy)
                                  {
                                    treecopy = CopyTree(treecopy,sl);
                                    pe->GetPE(1)->AddExpr(treecopy,1);
                                    if(substlist)
                                     { // Substitution in Baumkopie durchfuehren
                                      ParameterSubstitution(treecopy,substlist,sl);
                                      delete substlist;
                                     }
                                  }
                                 else
                                  {
                                    if(substlist)
                                     delete substlist;
                                  } 
                           }
                     
                           else
                              TreeImport(pe,start,sl);
                           break;
                                                         
        default  : assert(0);
        }   
      }
 }                    

ParseExpr* CopyTree(ParseExpr* expr,SymList* sl)
 { assert(expr && sl);
   int maxi =  expr->GetMaxIndex();
   noTerminal_Type nt = expr->GetNonTermType();
   char* rn = expr->GetRuleName(); 
   Symbol_Type st;
   ParseSymb* ps;
   SymListElem *se, *newse;
   ParseExpr* pe;
   

   ParseExpr* newexpr = new ParseExpr(nt, rn, maxi+1);


   for(int i = 1; i <= maxi+1 ; i++)
    {
      ps = expr->GetExpr(i);
      st = ps->GetSymbolType();
      switch(st)
       {
         case Empty : break;
                     
         case Terminal : se = ps->GetSymbElem();   // Terminalsymbol kopieren
                         if(se->tokenID == normalID)
                          { 
                            newse = se->copy();

                            sl->Append(newse); 
                            newexpr->AddExpr(newse ,i);
                          }
                         else
                            newexpr->AddExpr(se , i);

                         break;
                     
                         
         case NonTerm : pe = ps->GetParseExpr();  // Nichtterminalsymbol kopieren
                        newexpr->AddExpr(CopyTree(pe,sl) , i);
                        
                        break;
                     
         default : assert(0);
        }
     }
    return newexpr;
 }

void CorrectTree(ParseExpr* expr)
 {   assert(expr);
     int maxi = expr->GetMaxIndex();
     Symbol_Type st;    
     ParseSymb *ps;
     ParseExpr *pe,*pe1, *pe2, *pe3;
     noTerminal_Type tp;
     SymListElem* se;

     for(int i = 1 ; i<= maxi+1; i++)
      {
         ps = expr->GetExpr(i);
         st = ps->GetSymbolType();
         switch(st)
          {

            case Empty : break;

            case Terminal : break;

            case NonTerm : pe = ps->GetParseExpr();
                           tp = pe->GetNonTermType();
                           if ((tp == tID) &&    // ID mit Recordkomponentennamen
                               (pe->GetMaxIndex() == 1) &&
                               (pe->GetExpr(1)->GetSymbolType() == NonTerm) &&
                               (pe->GetExpr(1)->GetParseExpr()->GetNonTermType() == tqualifierPrefixes) &&
                               (pe->GetExpr(2)->GetSymbolType() == Terminal) &&
                               (pe->GetExpr(2)->GetSymbElem()->Get_nidt() == nidt_rec_comp))
                            { 

                              
                              pe1 = new ParseExpr(tID ,"ID", 2);
                              PopandAppend(pe->GetPE(1)->GetPE(1),pe1,1);
                              pe1->AddExpr(pe->GetPE(1)->GetSE(2),2);
                              pe2 = new ParseExpr(tcomponentID,"componentID",1);
                              pe2->AddExpr(pe->GetExpr(2)->GetSymbElem(),1); 
                              pe3 = new ParseExpr(corrected,"corrected",3);
                              PopandAppend(pe1,pe3,1);
                              pe3->AddExpr(pe->GetPE(1)->GetSE(3),2);
                              PopandAppend(pe2,pe3,3);
                              
                              PopandAppend(pe3,expr,i); 
                              
                             }
                            else 
                             CorrectTree(pe);                     
                            break;
             default : assert(0);
           }                              
        }
   }


void  ExamineSystemFormula(ParseExpr* tree,ParseExpr* start,SymList* sl)  //Bei ersten Aufruf: tree == start
 {  assert(tree && start && sl) ;
    int maxi = tree->GetMaxIndex();
    ParseExpr *pe;
    ParseSymb* ps;
    Symbol_Type st;
    noTerminal_Type nt;
    
    for(int i = 1 ; i<= maxi+1; i++)
     {
       ps = tree->GetExpr(i);
       st = ps->GetSymbolType();
       switch(st)
        {
          case Empty : break;
          case Terminal : break;
          case NonTerm : pe = ps->GetParseExpr();
                       nt = pe->GetNonTermType();
                if(  (nt == texpr)  &&                   //   Suche Konjunktion von Aktionen
                     (pe->GetMaxIndex() == 6) &&         // (Next State Relation)
                     (pe->GetExpr(1)->GetSymbolType() == Terminal) &&
                     (pe->GetSE(1)->tokenID == '[') &&
                     (pe->GetExpr(2)->GetSymbolType() == NonTerm) &&
                     (pe->GetPE(2)->GetNonTermType() == tActionExpr) &&
                     (pe->GetExpr(3)->GetSymbolType() == Terminal) &&
                     (pe->GetSE(3)->tokenID == ']')   )        
                   {  

                      ExpandExpression(pe->GetPE(2),start,sl);
                      pe_semantic_concept = pe->GetPE(2);


                      ExpandExpression(pe->GetPE(6),start,sl);


                      pe_variables = pe->GetPE(6);


                   }
                 else if(nt == tOneModule)  // Modulimporte brauchen hier nicht beruecksichtigt werden
                   { break;}
                 else
                   {  
                      ExamineSystemFormula(pe,start,sl);
                   }
                 break;
            default : assert(0);
          }
        }
    }                    
  



void  ExpandExpression(ParseExpr* expr,ParseExpr* start,SymList* sl)
 { assert(expr && start && sl);
   int maxi = expr->GetMaxIndex();
   ParseSymb* ps;
   ParseExpr *pe,*found,*expOp;
   SymListElem* se;
   Symbol_Type st;
   noTerminal_Type nt;
   char* opname;
   
   List *qualifiers, *actOpParams;
   for(int i = 1; i<= maxi+1; i++)
    {
      ps = expr->GetExpr(i);
      st = ps->GetSymbolType();
      switch(st)
       {
         case Empty : break;
         case Terminal : break;
         case NonTerm : pe = ps->GetParseExpr();
              nt = pe->GetNonTermType();
              switch(nt)
               {
                case tOperatorApplication :   // Expansion eines parametriesierten Bezeichners

                    qualifiers = new List;
                    actOpParams = new List;
                    se = pe->GetSE(2);
                    opname = se->Getlexem();

                    ExpandExpression(pe->GetPE(4),start,sl); // aktuelle Parameter expandieren

                    CollectQualifers(pe->GetPE(1),qualifiers, start,sl);
                    CollectActualParams(pe->GetPE(4),actOpParams); 

                    found = SearchforOperatorDef(start,start,opname,qualifiers,actOpParams,sl);
                              
                    delete qualifiers;
                    delete actOpParams;
                    if(found)
                     { 

                       delete pe;
                       PopandAppend(found, expr, i);
                     }
                    break;    

                case tSetExpr :            //  ID [ constExpr_csl]   

                    if( (pe->GetMaxIndex() == 3) &&
                        (pe->GetExpr(1)->GetSymbolType() == NonTerm) &&
                        (pe->GetPE(1)->GetNonTermType() == tID) &&
                        (pe->GetExpr(3)->GetSymbolType() == NonTerm) &&
                        (pe->GetPE(3)->GetNonTermType() == tconstExpr_csl) )
                       { 
                         se = pe->GetPE(1)->GetSE(2);
                         if( (se->Get_nidt() == nidt_operator) ||
                            (se->Get_nidt() == nidt_fctOperator) )
                           {

                             qualifiers = new List;
                             actOpParams = new List;
                             opname = se->Getlexem();


                             ExpandExpression(pe->GetPE(3),start,sl);

                             CollectQualifers(pe->GetPE(1)->GetPE(1),qualifiers, start,sl);
                             CollectActualParams(pe->GetPE(3),actOpParams);

                             found = SearchforOperatorDef(start,start,opname,qualifiers,actOpParams,sl);
                             delete qualifiers;
                             delete actOpParams;
                             if(found)
                             {

                               delete pe;
                               PopandAppend(found, expr, i);
                             }
                            }
                          }
                         else
                          {
                            ExpandExpression(pe,start,sl);
                          } 
                         break;

                case tID :  se = pe->GetSE(2);   // Expansion eines nich parametrisierten Bezeichners
                            if(    ( (se->Get_nidt() == nidt_operator) ||
                                     (se->Get_nidt() == nidt_fctOperator) ) &&
                                     (se->GetSort()->GetFirstElem()->GetSort() != s_const) )
                             {

                             qualifiers = new List;
                             actOpParams = new List;
                             opname = se->Getlexem();

                             CollectQualifers(pe->GetPE(1),qualifiers, start,sl);

                             found = SearchforOperatorDef(start,start,opname,qualifiers,actOpParams,sl);
                             delete qualifiers;
                             delete actOpParams;
                             if(found)
                               {

                                delete pe;
                                PopandAppend(found, expr, i);
                               }
                             }
                                                          
                           break;  
                default : ExpandExpression(pe,start,sl);
                          break;
               }
              break;
         default : assert(0);       
        }
      }
   }       



void CollectQualifers(ParseExpr* expr,List* qualList,ParseExpr* start,SymList* sl)
 { assert(expr&& qualList && start && sl);
   QualInfo* qi;
   IDinfo *id;
   List* modpars; 
   
   if(expr->GetMaxIndex() == 2)   // Qualifier ohne Modulparameter
     {
       CollectQualifers(expr->GetPE(1), qualList, start,sl);
       qualList->Append(new QualInfo(expr->GetSE(2)->Getlexem()));

     }
   else if(expr->GetMaxIndex() == 5)  // Qualifier mit Modulparameter
     {
        CollectQualifers(expr->GetPE(1), qualList, start,sl);      
        qi = new QualInfo(expr->GetSE(2)->Getlexem());


        ExpandExpression(expr->GetPE(4), start,sl);

        modpars = qi->GetQualParams();
        CollectActualParams(expr->GetPE(4), modpars);
        qualList->Append(qi);
     }
 }
  

void CollectActualParams(ParseExpr* expr,List* parList)
  { assert(expr && parList);
    IDinfo* id;
    if(expr->GetMaxIndex() == 0)
     {
        id = new IDinfo;
        id->SetExpression(expr->GetPE(1));
        parList->Append(id);
     }
    else
     {
        CollectActualParams(expr->GetPE(1),parList);
        id = new IDinfo;
        id->SetExpression(expr->GetPE(3));
        parList->Append(id); 
     }  

  }


ParseExpr* SearchforOperatorDef(ParseExpr* expr,ParseExpr* start,char* opname, List* qualList,List* parList,SymList* sl)
  { assert(start && opname && qualList && parList && sl);

    int maxi = expr->GetMaxIndex();
    int ok;
    ParseExpr *pe,*found;
    
    ParseSymb* ps;
    Symbol_Type st;
    SymListElem* se;
    noTerminal_Type nt;
    QualInfo *qi;
    List *modpars,*namelist;
    NameElem* nel;
    IDinfo* id;
    for(int i = 1 ; i<= maxi+1; i++)
     { 
       ps = expr->GetExpr(i);
       st = ps->GetSymbolType();
       switch(st)
       {
        case Empty : break;
        case Terminal : break;                             




        case NonTerm : pe = ps->GetParseExpr();
                       nt = pe->GetNonTermType();
            if((nt == tModuleName) && (pe->GetExpr(1)->GetSymbolType() == NonTerm))  // Importiertes Modul vorhanden
              {
                if (i  <  maxi+1)  // Qualifier ist vorhanden 
                   
                   { 
                      qi = (QualInfo*) qualList->GetFirstElem();
                      while(qi)
                       {
                         
                         if(strcmp(qi->GetQualName() , expr->GetPE(i+2)->GetSE(1)->Getlexem()) == 0)
                           {  
                              qi->SetPassed();   // Teilbaum wird besucht

                              if( (i+2 < maxi+1)  &&              // Modulparameter sind vorhanden
                                 (expr->GetPE(maxi+1)->GetMaxIndex() == 2) )
                               { modpars =  qi->GetQualParams();
                                 namelist = new List; 
                                 CollectFormalParameters_formalPar_ssl(expr->GetPE(maxi+1)->GetPE(2),namelist);
                                 id = (IDinfo*) modpars->GetFirstElem();
                                 nel = (NameElem*) namelist->GetFirstElem();
                                 while(id && nel) 
                                  {
                                    id->SetParname(nel->GetName());
                                    id = (IDinfo*) modpars->GetNextElem();
                                    nel = (NameElem*) namelist->GetNextElem();
                                  }
                                 delete namelist;

                               } 
                               found = SearchforOperatorDef(pe->GetPE(1),pe->GetPE(1), opname, qualList, parList,sl);
                               if(found)
                                return found;
                            }
                         qi = (QualInfo*) qualList->GetNextElem();
                       }
                      // gesuchter Operator besitzt diesen Qualifier nicht -> diesen Teilbaum nicht besuchen 
                      break;                         
                   }
                 else    // kein Qualifier, Suche in diesem Teilbaum fortsetzen
                   {
                     found = SearchforOperatorDef(pe,pe, opname, qualList, parList,sl);
                     if(found)
                        return found;
                   }
              }
            else if(nt == tSymbolDef_LHS)
             {
               ok = Examine_SymbolDef_LHS(pe,opname, parList,qualList);
               if(ok)  // Operator Definition gefunden
                 {  

                    found = CopyTree(expr->GetPE(i+2),sl);
                    
                    ParameterSubstitution(found,parList,sl);

                    ExpandExpression(found,found,sl);   // Suche von lokalen Operatordefinitionen


                    ExpandExpression(found,start,sl);

                    qi = (QualInfo*) qualList->GetFirstElem();
                    while(qi)
                     {
                       modpars = qi->GetQualParams();
                       if(modpars->GetLength() != 0)
                         ParameterSubstitution(found,modpars,sl);
                       qi = (QualInfo*) qualList->GetNextElem();
                     }
                   
                    return found;  
                       
                 }
               
              }
           
            else
            { 
              found = SearchforOperatorDef(pe,start, opname,qualList, parList,sl);
              if(found)
               return found;
              
             }
                   break;      
           default : assert(0);
        }
    }
   return NULL;
  }        
    
int passed_all_qualifiers(List* qL)   // Sind Qualifier des Bezeichners bei der Suche beruecksichtigt worden?
 { assert(qL);
   if(qL->GetLength() == 0)
     return 1;
   QualInfo* qi = (QualInfo*) qL->GetFirstElem();
   while(qi)
     {
      if( !(qi->isPassed()))
        {

         return 0;
        } 
      else
        {
        qi = (QualInfo*) qL->GetNextElem();
        }
     }
   return 1; 
 }  

int Examine_SymbolDef_LHS(ParseExpr* expr, char* opname, List* parList,List* qualList)
  {  assert(expr && opname && parList && qualList);

     int maxi = expr->GetMaxIndex();
     List* params;
     NameElem* nel;
     IDinfo* id;


     if((maxi == 1) && passed_all_qualifiers(qualList))
      {
        if(strcmp(opname, expr->GetSE(1)->Getlexem()) == 0) // Operatordefinition gefunden
         {

           return 1;
         }
        else 
          return 0;  // nicht gefunden
      }
     else if((maxi == 4) && passed_all_qualifiers(qualList))
      {
         if(strcmp(opname, expr->GetSE(1)->Getlexem()) == 0) // Operatordefinition gefunden     
            { 

              params = new List;
              CollectFormalParameters_formalPar_ssl(expr->GetPE(3),params);
              nel = (NameElem*) params->GetFirstElem();
              id = (IDinfo*) parList->GetFirstElem();
              while(nel && id)
               {
                 id->SetParname(nel->GetName());
                 nel = (NameElem*) params->GetNextElem();
                 id = (IDinfo*) parList->GetNextElem();
               }
              delete params;    
              return 1;
            }
         else
           return 0;
      }
     else if((maxi == 8) && passed_all_qualifiers(qualList))
       {
         if(strcmp(opname, expr->GetSE(5)->Getlexem()) == 0) // Operatordefinition gefunden    
          {

            params = new List;
            CollectFormalParameters_formalPar(expr->GetPE(3),params);
            CollectFormalParameters_formalPar(expr->GetPE(7),params);
            nel = (NameElem*) params->GetFirstElem();
            id = (IDinfo*) parList->GetFirstElem();
            while(nel && id)
             {
               id->SetParname(nel->GetName());
               nel = (NameElem*) params->GetNextElem();
               id = (IDinfo*) parList->GetNextElem();
             }
            delete params;

            return 1;
          }
         else
            return 0;
      }
  return 0;
  }  






void CollectFormalParameters_formalPar_ssl(ParseExpr* expr,List* parList)
 { assert(expr && parList);

   if(expr->GetMaxIndex() == 2)
    {
      CollectFormalParameters_formalPar_ssl(expr->GetPE(1),parList);

      CollectFormalParameters_formalPar(expr->GetPE(3),parList);
    }
   else
    {
       CollectFormalParameters_formalPar(expr->GetPE(1),parList);
    }
 }
   

void CollectFormalParameters_formalPar(ParseExpr* expr, List* parList)
 { assert(expr && parList);
   if(expr->GetMaxIndex() == 1)
     {
       CollectFormalParameters_normalID_csl(expr->GetPE(1),parList);
     }
   else if(expr->GetMaxIndex() == 4)
     {
       parList->Append(new NameElem(expr->GetSE(1)->Getlexem()));
     }
   else
     {
       parList->Append(new NameElem(expr->GetSE(5)->Getlexem()));
     }
 }
 

void CollectFormalParameters_normalID_csl(ParseExpr* expr, List* parList)
  { assert(expr&& parList);
   
    if(expr->GetMaxIndex() == 2)
      {
        CollectFormalParameters_normalID_csl(expr->GetPE(1),parList);
        parList->Append(new NameElem(expr->GetSE(3)->Getlexem()));
      }
    else
      {
        parList->Append(new NameElem(expr->GetSE(1)->Getlexem()));
      }
   }



int isUNCHANGEDconstruct(ParseExpr* expr)
 { assert(expr);
  
   if(expr->GetPE(2)->GetSE(1)->tokenID == EQ) 
        
     {
        return CompareExprs(expr->GetPE(1), expr->GetPE(3));

     }
   else
    return 0;
 }


int CompareExprs(ParseExpr* expr1, ParseExpr* expr2)
 { assert(expr1 && expr2);
   ParseExpr *pe1, *pe2;
   ParseSymb *ps1, *ps2;
   SymListElem *se1, *se2;
   Symbol_Type st1, st2;
   noTerminal_Type nt1, nt2;
   int equal =1 ;

   int maxi1 = expr1->GetMaxIndex() , maxi2 = expr2->GetMaxIndex(); 

   if( (maxi1 == 1) && (maxi2 == 0) &&
       (expr1->GetExpr(1)->GetSymbolType() == NonTerm) &&
       (expr1->GetExpr(2)->GetSymbolType() == Terminal) &&
       (expr1->GetSE(2)->tokenID == PRIME) )
        { 
          return CompareExprs(expr1->GetPE(1), expr2);
        }
   else if( (maxi2 == 1) && (maxi1 == 0) &&
            (expr2->GetExpr(1)->GetSymbolType() == NonTerm) &&
            (expr2->GetExpr(2)->GetSymbolType() == Terminal) &&
            (expr2->GetSE(2)->tokenID == PRIME) )
        {
          return CompareExprs(expr1, expr2->GetPE(1));
        }   

   else if(maxi1 == maxi2)
     { 
       for(int i =1 ; i <= maxi1 + 1; i++)
         {
           ps1 = expr1->GetExpr(i);
           ps2 = expr2->GetExpr(i);
           st1 = ps1->GetSymbolType();
           st2 = ps2->GetSymbolType();
           if(st1 == st2)
             {
              switch(st1)
               {
                case Empty : break;
                case Terminal : se1 = ps1->GetSymbElem();
                                se2 = ps2->GetSymbElem();
                                if(se1->tokenID == se2->tokenID)
                                  {
                                    if( (se1->tokenID == normalID) &&
                                        (strcmp(se1->Getlexem(),se2->Getlexem()) != 0) )
                                     return 0; //  equal = 0;
                                    else
                                       equal =1;  
                                   }
                                  else 
                                    return 0;   // equal = 0;
                                  break;            
                  case NonTerm : pe1 = ps1->GetParseExpr();
                                 pe2 = ps2->GetParseExpr();
                                 if(pe1->GetNonTermType() == pe2->GetNonTermType())
                                  {
                                     equal = equal && CompareExprs(pe1,pe2);
                                     
                                  }
                                 else 
                                   return 0;  // equal = 0;
                                 break;
                  default : assert(0);
                  } // end of switch
                }   // end of if
            else
              return 0; //equal = 0;
      } // end of for
      return equal; 
     }
  else 
    return 0; // return 0; 
   }


   

void ExamineTerm(ParseExpr* expr,AdjList* SCadjl, int actionnr)
  { assert(expr && actionnr);
    int lfoundvar = 0, rfoundvar = 0;
    int lopenbraces = 0, ropenbraces = 0;
   
    if(! isUNCHANGEDconstruct(expr))
      { 

         symbols = new SymList;
         VariableAccessSearch(expr->GetPE(1), &lfoundvar,&lopenbraces);
         if(symbols->GetLength() > 0)
           AppendVariablesToAdjList(SCadjl,actionnr);
         delete symbols;
         symbols = NULL;
         symbols = new SymList;

         VariableAccessSearch(expr->GetPE(3),&rfoundvar,&ropenbraces);
         if(symbols->GetLength() > 0)
          AppendVariablesToAdjList(SCadjl,actionnr);
         delete symbols;
        symbols = NULL;
      }

  }    

void VariableAccessSearch(ParseExpr* expr,  int* foundvar,int* openbraces)  // Ermittlung und Speicherung
                                                                            // der benutzten Systemvariablen bzw.
                                                                            // Variablenkomponenten 
 {  assert(expr );
    int tid,maxi = expr->GetMaxIndex();
    ParseExpr* pe;
    ParseSymb* ps;
    SymListElem* se;
    Symbol_Type st;
    noTerminal_Type nt;
   
    for(int i =1 ; i<= maxi+1;i++)
     {
       ps = expr->GetExpr(i);
       st = ps->GetSymbolType();
       switch(st)
        {
          case Empty : break;
          case Terminal : if(*foundvar)
                           {se = ps->GetSymbElem();
                            tid = se->tokenID;
                             switch(tid)
                              {
                              case normalID :  
                                               symbols->Append(se); break;
                              case '[' : *openbraces = *openbraces + 1; symbols->Append(se); break;
                              case ']' : if(*openbraces == 0)
                                         { 
                                           
                                           *foundvar = 0;  break;
                                          }
                                         else
                                          {*openbraces = *openbraces-1;
                                           symbols->Append(se); break;
                                          };
                              case RBR_ON : if(*openbraces)
                                            {*openbraces = *openbraces + 1;
                                              symbols->Append(se);}
                                             else 
                                               *foundvar = 0; 
                                             break;
                              case RBR_OFF : if(*openbraces == 0)
                                         { 
                                            
                                            *foundvar = 0;  break;
                                          }
                                         else
                                          {*openbraces = *openbraces-1;
                                           symbols->Append(se); break;
                                          };
                              case TUPELSELECTOR : symbols->Append(se); break;
                              case DOT : symbols->Append(se); break;
                              case NATNUM : if(*openbraces) symbols->Append(se);
                                            else 
                                              *foundvar = 0; 
                                             break;
                              case STRING : if(*openbraces) symbols->Append(se);
                                            else
                                              *foundvar = 0;
                                             break;
                              case PLUS : if(*openbraces)
                                           symbols->Append(se);
                                          else 
                                             *foundvar = 0;
                                           break;
                              case MINUS : if(*openbraces)
                                           symbols->Append(se);
                                          else
                                             *foundvar = 0;
                                           break; 
                              case TIMES : if(*openbraces)
                                           symbols->Append(se);
                                          else
                                             *foundvar = 0;
                                           break;
                              case DIVIDED : if(*openbraces)
                                           symbols->Append(se);
                                          else
                                             *foundvar = 0;
                                           break;
                              case DIV :  if(*openbraces)
                                           symbols->Append(se);
                                          else
                                             *foundvar = 0;
                                           break;
                              case MOD : if(*openbraces)
                                           symbols->Append(se);
                                          else
                                             *foundvar = 0;
                                           break;
                              case POWER : if(*openbraces)
                                           symbols->Append(se);
                                          else
                                             *foundvar = 0;
                                           break;
                              case COMMA : if(*openbraces)
                                           symbols->Append(se);
                                          else
                                             *foundvar = 0;
                                           break;
                                          
                              default : 
                                        *foundvar = 0; break;
                              }
                            }
                            
         
                         
                           break;
          case NonTerm :
             pe = ps->GetParseExpr();
             nt =pe->GetNonTermType();

             if((nt == tID) && !(*foundvar))
              { se = pe->GetSE(2);
                // Systemvariable muss die Sorte STATE FUNCTION besitzen
                if(se->GetSort()->GetFirstElem()->GetSort() == s_statefct )
                 {

                    *foundvar = 1; 
                    *openbraces = 0; 
                    
                    
                    VariableAccessSearch(pe,  foundvar,openbraces);

                 }
                else
                 VariableAccessSearch(pe,foundvar,openbraces);
              }
             else
              VariableAccessSearch(pe,foundvar,openbraces);
             break;
          default : assert(0);
          }
      }
  }   

void CreateSemanticConcept(ParseExpr* varexpr,ParseExpr* actionexpr,AdjList* SCadjl)
 { 
   Node *top,*n;
   int foundvar = 0,openbraces = 0;
   if(varexpr && actionexpr)
    {

   SCadjl->SCAddNode("_STATE_",0,'X');    // Wurzel des Variablenbaums
   symbols = new SymList;     
   VariableAccessSearch(varexpr,  &foundvar, &openbraces);
    // Alle Systemvariablen in der Variablenliste der Formel werden beruecksichtigt
   if(symbols->GetLength() > 0) 
     AppendVariablesToAdjList(SCadjl,0);

   delete symbols;
   symbols = NULL;
    
   top = SCadjl->GetFirstNode();
   n = SCadjl->GetNextNode();
     while(n)
      {

        top->AddEdge(n,'x');
        n = SCadjl->GetNextNode();
      }
   delete symbols; 

  ExamineOperation(actionexpr,SCadjl,OR_Level, 0);

  SCAddEdgesToNodes(SCadjl);

    } 

 }
        
void ExamineOperation(ParseExpr* expr,AdjList* SCadjl, Node_Level nl, int actionnr)
 {
   assert(expr);
   int newactionnr, maxi = expr->GetMaxIndex();
   ParseExpr* pe;
   ParseSymb* ps;
   SymListElem *se;
   Symbol_Type st;
   noTerminal_Type nt;

   for(int i = 1; i<= maxi+1;i++)
    {
      ps = expr->GetExpr(i);
      st = ps->GetSymbolType();
      switch(st)
       {
        case Empty : break;
        case Terminal : break;
        case NonTerm : pe = ps->GetParseExpr();
                        nt = pe->GetNonTermType();
                        if( (nt == texpr) &&            
                            (pe->GetMaxIndex() == 2) &&
                            (pe->GetExpr(2)->GetSymbolType() == NonTerm) &&
                            (pe->GetPE(2)->GetNonTermType() == tboolRel) )
                         {  // ODER Ebene
                           if(pe->GetPE(2)->GetSE(1)->tokenID == _L_OR_ )
                             {

                               ExamineOperation(pe,SCadjl, nl, actionnr);
                              
                               break;
                             }
                           else if(nl == OR_Level)  // Uebergang ODER ->UND Ebene
                             { 
                               ActionNr++;  // Neue Aktion beginnt
                               newactionnr = ActionNr;

                               ExamineOperation(pe,SCadjl, AND_Level, newactionnr);
                               
                               break;
                             }
                           else
                              { 
                                ExamineOperation(pe,SCadjl, nl, actionnr);

                                break;
                              }  
                         }
                        else if( (nt == texpr) &&            // Praedikat gefunden
                                 (pe->GetMaxIndex() == 2) &&
                                 (pe->GetExpr(2)->GetSymbolType() == NonTerm) &&
                                 (pe->GetPE(2)->GetNonTermType() == tBinRel) )
                                {
                                   if(nl == OR_Level)
                                    {
                                      ActionNr++;
                                      newactionnr = ActionNr;

                                      ExamineTerm(pe,SCadjl, newactionnr);
                                      break; 
                                    }
                                   else                    
                                    {

                                      ExamineTerm(pe,SCadjl,actionnr);
                                      break;
                                    }
                                 }
                          else if(nt == tconjunctLines)
                           {  
                              if(nl == OR_Level)   // ODER-Level  -> UND-Level   
                               {
                                 ActionNr++;  // Aktionsnummer erhoehen
                                 newactionnr = ActionNr;

                                 ExamineOperation(pe,SCadjl,AND_Level, newactionnr);
                                 break;
                               }
                              else
                               {

                                 ExamineOperation(pe,SCadjl,nl, actionnr);  
                                 break;
                               }                                

                           }
                          else if(nt == tdisjunctLines)
                           {

                             ExamineOperation(pe,SCadjl,nl,actionnr);
                             break;
                           }
                          else
                            {
                              ExamineOperation(pe,SCadjl, nl, actionnr);
                              break;
                            }
                    default : assert(0);
                    } // end switch
                } // end For
           } //end ExamineOperation    
                                
void AppendVariablesToAdjList(AdjList* SCadjl,int actnr)
 { 
   assert(SCadjl && symbols);
   SymListElem* se;
   SymListElem **SymElemArray;
   char *varname = NULL,*help;

   int isarray ,openbr,tid,count,var,NrOfSymbols = symbols->GetLength();
      
   SymElemArray = new SymListElem*[NrOfSymbols];
   count = 0;
   se = symbols->GetFirstElem();

   while(se && (count < NrOfSymbols))
    {


      SymElemArray[count] = se;
      count++;
      se = symbols->GetNextElem();
    }

   var = 0;
   
   for(int i = 0; i <= NrOfSymbols-1; i++)
   {
     if(varname) 
      {delete varname;varname = NULL;}
     if( (SymElemArray[i]->tokenID == normalID ) &&
         (SymElemArray[i]->Get_nidt() != nidt_rec_comp) &&
         (SymElemArray[i]->GetSort()->GetFirstElem()->GetSort() == s_statefct) ) 
       {  
          isarray = 0; 
          varname = new char[strlen(SymElemArray[i]->Getlexem()) + 1];
          strcpy(varname, SymElemArray[i]->Getlexem());



          openbr = 0;

       }
      else 
       continue;
      
     for(int j = i+1; j <= NrOfSymbols-1; j++)
      {
         tid = SymElemArray[j]->tokenID;   // Erzeugung der korrekten Variablen-Strings
                                           // Evtl. mit Komponentennamen. 
         switch(tid)
          {
           case normalID :  if((openbr > 0) || ((openbr == 0) &&(SymElemArray[j-1]->tokenID == DOT)) ) 
                             {
                                help = varname;
                                varname = new char[strlen(SymElemArray[j]->Getlexem()) + 
                                               strlen(help) + 1];
                                strcpy(varname, help);
                                strcat(varname, SymElemArray[j]->Getlexem());
                                delete help;
                                
                             }
                             else
                              openbr = -1; 
                             break; 
                      
           case '[' : if(openbr < 0) break;
                      if((openbr == 0) && (isarray))
                         SCadjl->SCAddNode(varname, 0,'A');
                      else if((openbr == 0) && (! isarray))
                         {
                           SCadjl->SCAddNode(varname, 0,'X');
                           isarray = 1;
                         }
                      openbr++; help = varname;
                      varname = new char[strlen("[") +
                                               strlen(help) + 1];
                      strcpy(varname, help);
                      strcat(varname,"[" );                              
                      delete help;
                      break;
                      
           case ']' : if(openbr <= 0) break;
                      openbr--;
                      help = varname;
                      varname = new char[strlen("]") +
                                               strlen(help) + 1];
                      strcpy(varname, help);
                      strcat(varname,"]" );
                      delete help;
                      if(openbr == 0)
                       {

                          SCadjl->SCAddNode(varname, 0,'A');
                       } 
                      break;   
                  
           case  RBR_ON :if(openbr <= 0) break; 
                         if(openbr)
                           {
                            openbr++; help = varname;
                            varname = new char[strlen("(") +
                                               strlen(help) + 1];             
                            strcpy(varname, help);
                            strcat(varname,"(" );
                            delete help;
                           }
                          break;
           case RBR_OFF :  if(openbr <= 0) break;
                           openbr--;
                           
                            help = varname;
                            varname = new char[strlen(")") +
                                               strlen(help) + 1];
                            strcpy(varname, help);
                            strcat(varname,"(" );
                            delete help;
                           
                            break;
           case TUPELSELECTOR : if(openbr < 0) break;
                                if((openbr == 0) && (!isarray))
                                 SCadjl->SCAddNode(varname, 0,'X');
                                else if((openbr == 0) && (isarray))
                                 SCadjl->SCAddNode(varname, 0,'A'); 
                                help = varname;
                                varname = new char[strlen(SymElemArray[j]->Getlexem()) +
                                                   strlen(help) + 1];
                                strcpy(varname, help);
                                strcat(varname,SymElemArray[j]->Getlexem());
                                delete help;
                                break;
            case DOT : if(openbr < 0) break;
                       if((openbr == 0) && (!isarray))
                         SCadjl->SCAddNode(varname, 0,'X');
                       else if((openbr == 0) && (isarray))
                         SCadjl->SCAddNode(varname, 0,'A'); 
                       help = varname;
                       varname = new char[strlen(".") + strlen(help) + 1];
                       strcpy(varname, help);
                       strcat(varname,".");
                       delete help;
                       break;
            case NATNUM : if(openbr <= 0)  break;
                          help = varname;
                          varname = new char[strlen(SymElemArray[j]->Getlexem()) +
                                                  strlen(help) + 1];
                          strcpy(varname, help);
                          strcat(varname,SymElemArray[j]->Getlexem());
                          delete help;
                                               
                          break; 
            case STRING : if(openbr <= 0)  break;
                          help = varname;
                          varname = new char[strlen(SymElemArray[j]->Getlexem()) +
                                                  strlen(help) + 1];
                          strcpy(varname, help);
                          strcat(varname,SymElemArray[j]->Getlexem());
                          delete help;

                          break;
            case PLUS  : if(openbr <= 0) break;
                         
                            help = varname;
                            varname = new char[strlen("+") + strlen(help) + 1];
                            strcpy(varname, help);
                            strcat(varname,"+");
                            delete help;
                            break;
                      
            case MINUS : if(openbr <=0) break;;
                        
                            help = varname;
                            varname = new char[strlen("-") + strlen(help) + 1];
                            strcpy(varname, help);
                            strcat(varname,"-");
                            delete help;
                            break;
            case TIMES : if(openbr <=0 ) break;
                         
                            help = varname;
                            varname = new char[strlen("*") + strlen(help) + 1];
                            strcpy(varname, help);
                            strcat(varname,"*");
                            delete help;
                            break;
                         
            case DIVIDED : if(openbr <= 0) break;
                        
                            help = varname;
                            varname = new char[strlen("/") + strlen(help) + 1];
                            strcpy(varname, help);
                            strcat(varname,"/");
                            delete help;
                            break;
                       
             case DIV : if(openbr <= 0) break;
                       
                            help = varname;
                            varname = new char[strlen("DIV") + strlen(help) + 1];
                            strcpy(varname, help);
                            strcat(varname,"DIV");
                            delete help;
                            break;
                       
              case MOD : if(openbr <= 0) break;
                        
                            help = varname;
                            varname = new char[strlen("MOD") + strlen(help) + 1];
                            strcpy(varname, help);
                            strcat(varname,"MOD");
                            delete help;
                            break;
                        
              case POWER : if(openbr <= 0) break;
                        
                            help = varname;
                            varname = new char[strlen("^") + strlen(help) + 1];
                            strcpy(varname, help);
                            strcat(varname,"^");
                            delete help;
                            break;
               case COMMA : if(openbr <= 0) break;
                            help = varname;
                            varname = new char[strlen(",") + strlen(help) + 1];
                            strcpy(varname, help);
                            strcat(varname,",");
                            delete help;
                            break;
              
               default : assert(0);break;
               }
          }
       if(! isarray)
        SCadjl->SCAddNode(varname,actnr,'X');     
       else
        SCadjl->SCAddNode(varname,actnr,'A');
       }

   delete [NrOfSymbols]  SymElemArray;
  }
