/*----------------------------------------------------------------------*
 * Sort.c
 *
 * Michael Geikowski, 5.6.93
 *
 * Enthaelt die Member-Funktionen der Klassen SortListElem und Sort
 * sowie einige der Klasse Sort verwandte Non-Member-Funktionen zur
 * Unterstuetzung der Sortenueberpruefung durch den Parser
 *----------------------------------------------------------------------*/

#include <stdio.h>
#include <assert.h>

#include "sort.h"
#include "dbinterface.h"

/*----------------------------------------------------------------------*
 * Klasse SortListElem
 *----------------------------------------------------------------------*/

SortListElem::~SortListElem()
 {
 assert( ((nested==0) && (the_Sort!=s_nested)) ||
         ((nested!=0) && (the_Sort==s_nested))  );
 if (nested) delete nested;			// evtl. loeschen
 }

/*----------------------------------------------------------------------*
 * Klasse Sort
 *----------------------------------------------------------------------*/

int Sort::operator==(Sort_Type t)
 {
 SortListElem *p;

 p = (SortListElem*) List::GetFirstElem();
 return (p != 0) && 				// Zeigt this auf ein-
        (p->the_Sort == t) && 			// elementige Liste mit Element
        (List::GetNextElem()==0);		// mit Sorten-Flag t ?
 }

static int primMatch(Sort_Type t0, Sort_Type t)
	// Hilfsfunktion: ergibt True, wenn t0 und t Sorten-Flags fuer primitive
        // Sorten sind und t0 gemaess der Halbordnung ueber den primitiven
        // Sorten kleiner oder gleich t ist.
 {
 if (t==s_joker) return 1;
 switch(t0)
   {
   case s_const:    return (t==s_const)||(t==s_statefct)||(t==s_transfct);
   case s_statefct: return               (t==s_statefct)||(t==s_transfct);
   case s_transfct: return                                (t==s_transfct);
   case s_bool:     return (t==s_bool)||
                           (t==s_pred)||(t==s_action)||(t==s_temporal);
   case s_pred:     return (t==s_pred)||(t==s_action)||(t==s_temporal);
   case s_action:   return              (t==s_action)||(t==s_temporal);
   case s_temporal: return                             (t==s_temporal);
   case s_joker:    return 1;
   default: return 0;
   };
 }

int Sort::operator<=(Sort_Type t)
 {
 SortListElem *p = (SortListElem*) List::GetFirstElem();
 assert(p);
 if (List::GetNextElem())
   {
   return 0;
   }
 return primMatch(p->the_Sort,t);
 }

int Sort::operator<=(Sort& s)
 {
 SortListElem *p,*q;			// SortListElems von 'this' bzw. 's'

 p = (SortListElem*)   List::GetFirstElem();
 q = (SortListElem*) s.List::GetFirstElem();
 while (p && q)				// Solange keine Liste am Ende
  {
  if ( ( p->the_Sort == s_nested ) &&
       ( q->the_Sort == s_nested )  )	// Wenn beide Sorten 'nested'...
    {
    if (! (*(p->nested) <= *(q->nested)))// Vergleiche rekursiv
      {
      return 0;
      }
    }
  else if (! primMatch(p->the_Sort,q->the_Sort)) // Vergleiche primitive
    {						 // Sorten
    return 0;
    }
  p = (SortListElem*)   List::GetNextElem();
  q = (SortListElem*) s.List::GetNextElem();
  }
 return (p==0) && (q==0);		// legal, wenn beide gleich lang
 }

int Sort::legal_call(Sort& s)
 {
 SortListElem *p,*q;			// SortListElems von 'this' bzw. 's'

 p = (SortListElem*)   List::GetFirstElem();
 q = (SortListElem*) s.List::GetFirstElem();
 while (p && q)				// Solange keine Liste am Ende
  {
  if ( ( p->the_Sort == s_nested ) &&
       ( q->the_Sort == s_nested )  )	// Wenn beide Sorten 'nested'...
    {
    if (! (*(q->nested) <= *(p->nested))) // Vergleiche rekursiv
      {
      return 0;
      }
    }
  else if (! primMatch(q->the_Sort,p->the_Sort)) // Vergleiche primitive
    {						 // Sorten
    return 0;
    }
  p = (SortListElem*)   List::GetNextElem();
  q = (SortListElem*) s.List::GetNextElem();
  }
 return (p) && (List::GetNextElem()==0);// legal, wenn p um eins laenger
 }

Sort* Sort::copy()
 {
 Sort         *s = new Sort;
 SortListElem *p;

 p = (SortListElem*) List::GetFirstElem();
 while (p)
  {
  if ( p->the_Sort == s_nested )	// wenn nested kopiere rekursiv
    {
    s->Append(new SortListElem(s_nested,p->nested->copy()));
    }
  else					// sonst haenge einfach primitive
    {					// Sorte an
    s->Append(p);
    }
  p = (SortListElem*) List::GetNextElem();
  }
 return s;
 }

Sort* Sort::raise(int level)
 {
 Sort         *s = new Sort;
 SortListElem *p,*q;

 assert(1<=level && level<=3);
 p = (SortListElem*) List::GetFirstElem();
 while (p)
  {
  if ( p->the_Sort == s_const )		// anheben CONSTANT -> ...
    {
    switch(level)
      {
      case 1: s->Append(new SortListElem(s_statefct)); break;
      case 2: s->Append(new SortListElem(s_transfct)); break;
      case 3: s->Append(new SortListElem(s_transfct)); break;
      default: assert(0);
      }
    }
  else if ( p->the_Sort == s_bool )	// anheben BOOLEAN -> ...
    {
    switch(level)
      {
      case 1: s->Append(new SortListElem(s_pred)); break;
      case 2: s->Append(new SortListElem(s_action)); break;
      case 3: s->Append(new SortListElem(s_temporal)); break;
      default: assert(0);
      }
    }
  else					// kann nicht angehoben werden
    {
    break;
    }
  p = (SortListElem*) List::GetNextElem();
  }
 q = (SortListElem*) List::GetLastElem();
 if (p ||
     ((level==3) && (q->the_Sort != s_bool)) ||
     (List::GetLength()<2) )
   {
   delete s;				// wenn anheben schiefgelaufen
   s=0;
   }
 return s;
 }

static void PrimSortErrorOutPut(Sort_Type t)
	// Hilfsfunktion zu ErrorOutPut. Gibt ASCII Text zu Sorten-Flag t aus.
 {
 switch(t)
   {
   case s_const:    SimpleErrorCont("CONSTANT");  break;
   case s_bool:     SimpleErrorCont("BOOLEAN");   break;
   case s_statefct: SimpleErrorCont("STATE FCN"); break;
   case s_pred:     SimpleErrorCont("PREDICATE"); break;
   case s_transfct: SimpleErrorCont("TRANS FCN"); break;
   case s_action:   SimpleErrorCont("ACTION");    break;
   case s_temporal: SimpleErrorCont("TEMPORAL");  break;
   default:         SimpleErrorCont("UNKNOWN");   break;
   };
 }

void Sort::ErrorOutPut(int isParList)
 {
 SortListElem *p;

 int l = GetLength();
 int i = 0;
 p = (SortListElem*) List::GetFirstElem();
 while (p)
  {
  if ( p->the_Sort == s_nested )
    {
    SimpleErrorCont("(");		// rekursive Ausgabe
    p->nested->ErrorOutPut();
    SimpleErrorCont(")");
    }
  else
    {
    PrimSortErrorOutPut(p->the_Sort);	// einfache Ausgabe
    }
  i++;
  if ( (i==l-1) && (! isParList) ) SimpleErrorCont(" -> ");
  else if (i < l) SimpleErrorCont(" x ");
  p = (SortListElem*) List::GetNextElem();
  }
 }

void Sort::ErrorOutPut2()
 {
 SortListElem *p;

 int l = GetLength();
 int i = 1;
 p = (SortListElem*) List::GetFirstElem();
 while (i<l)
  {
  if ( p->the_Sort == s_nested )
    {
    SimpleErrorCont("(");
    p->nested->ErrorOutPut();		// zusammengesetzte Sorte ausgeben
    SimpleErrorCont(")");
    }
  else
    {
    PrimSortErrorOutPut(p->the_Sort);	// primitive Sorte ausgeben
    }
  i++;
  if (i < l) SimpleErrorCont(" x ");
  p = (SortListElem*) List::GetNextElem();
  }
 }

/*----------------------------------------------------------------------*
 * non-member-functions
 *----------------------------------------------------------------------*/

Sort* BinaryMaxMatch(Sort *s1, Sort *s2, Sort_Type t,
                     long errpos, int errleng, Sort_Type def)
 { 
 assert(s1 && s2);
 if ( (*s1 <= t) && (*s2 <= t) )
   {
   if (*s1 <= (s2)->GetFirstElem()->GetSort())
     {
     delete s1; return s2;
     }
   else
     {
     delete s2; return s1;
     }
   }
 else
   {
   if (*s1 <= t)
     {
     SimpleError("sort mismatch (last argument)",errpos,errleng);
     SimpleErrorCont("  used     : ");
     s2->ErrorOutPut();
     }
   else
     {
     SimpleError("sort mismatch (first argument)",errpos,errleng);
     SimpleErrorCont("  used     : ");
     s1->ErrorOutPut();
     }
   SimpleErrorCont("\n  expected : ");
   PrimSortErrorOutPut(t);
   SimpleErrorCont(" (or subsort)\n");
   delete s1; delete s2;
   return new Sort (def);
   }
 }

Sort* UnaryMaxMatch(Sort *s, Sort_Type t,
                    long errpos, int errleng, Sort_Type def)
 { 
 assert(s);
 if (*s <= t)
   {
   return s;
   }
 else
   {
   SimpleError("sort mismatch",errpos,errleng);
   SimpleErrorCont("  used     : ");
   s->ErrorOutPut();
   SimpleErrorCont("\n  expected : ");
   PrimSortErrorOutPut(t);
   SimpleErrorCont(" (or subsort)\n");
   delete s;
   return new Sort (def);
   }
 }

void UnaryExactMatch(Sort *s, Sort_Type t, long errpos, int errleng)
 { 
 assert(s);
 if (! (*s == t) )
   {
   SimpleError("sort mismatch",errpos,errleng);
   SimpleErrorCont("  used     : ");
   s->ErrorOutPut();
   SimpleErrorCont("\n  expected : ");
   PrimSortErrorOutPut(t);
   SimpleErrorCont("\n");
   }
 }

Sort* BinaryMaxBoolMatch(Sort *s1, Sort *s2, long errpos, int errleng)
 {
 Sort_Type constSort, boolSort = s_joker;

 assert(s1 && s2);
 Sort *s = BinaryMaxMatch(s1,s2,s_transfct,errpos,errleng,s_const);
 constSort = s->GetFirstElem()->GetSort();
 if (constSort == s_const)    boolSort = s_bool;
 if (constSort == s_statefct) boolSort = s_pred;
 if (constSort == s_transfct) boolSort = s_action;
 delete s;
 return new Sort(boolSort);
 }

Sort* BoolConstToConst(Sort *boolSort, Sort *constSort)
 {
 Sort_Type primConstSort1, primConstSort2, primBoolSort;

 assert(boolSort && constSort);
 primBoolSort   = boolSort->GetFirstElem()->GetSort();
 primConstSort2 = constSort->GetFirstElem()->GetSort();
 delete boolSort;
      if (primBoolSort == s_bool)   primConstSort1 = s_const;
 else if (primBoolSort == s_pred)   primConstSort1 = s_statefct;
 else if (primBoolSort == s_action) primConstSort1 = s_transfct;
 else primConstSort1 = s_joker;
 if (primMatch(primConstSort1,primConstSort2))
   {
   return constSort;
   }
 else if (primMatch(primConstSort2,primConstSort1))
   {
   delete constSort;
   return new Sort(primConstSort1);
   }
 else assert(0);
 return(0);
 }

Sort* BoolConstToBool(Sort *boolSort, Sort *constSort)
 {
 Sort_Type primConstSort, primBoolSort1, primBoolSort2;

 assert(boolSort && constSort);
 primBoolSort1 = boolSort->GetFirstElem()->GetSort();
 primConstSort = constSort->GetFirstElem()->GetSort();
 delete constSort;
 if (primBoolSort1 == s_temporal)
   {
   return boolSort;
   }
      if (primConstSort == s_const)    primBoolSort2 = s_bool;
 else if (primConstSort == s_statefct) primBoolSort2 = s_pred;
 else if (primConstSort == s_transfct) primBoolSort2 = s_action;
 else primBoolSort2 = s_joker;
 if (primMatch(primBoolSort2,primBoolSort1))
   {
   return boolSort;
   }
 else if (primMatch(primBoolSort1,primBoolSort2))
   {
   delete boolSort;
   return new Sort(primBoolSort2);
   }
 else assert(0);
 return(0);
}


Sort* MaxMatch(Sort *s1, Sort *s2)
 {
 assert(s1 && s2);
 if (*(s1) <= (s2)->GetFirstElem()->GetSort())
   {
   delete s1;
   return s2;
   }
 else if (*(s2) <= (s1)->GetFirstElem()->GetSort())
   {
   delete s2;
   return s1;
   }
 else
  {
  assert(0);
  return(0);
  }
 }

Sort* MaxBoolMatch(Sort *s1, Sort *s2)
 {
 Sort_Type constSort, boolSort = s_joker;

 assert(s1 && s2);
 Sort *s = MaxMatch(s1,s2);
 constSort = s->GetFirstElem()->GetSort();
 if (constSort == s_const)    boolSort = s_bool;
 if (constSort == s_statefct) boolSort = s_pred;
 if (constSort == s_transfct) boolSort = s_action;
 delete s;
 return new Sort(boolSort);
}

Sort* MaxConstMatchOfMixedSorts(Sort *boolSort, Sort *constSort)
 {
 Sort_Type primConstSort, primBoolSort;

 assert(boolSort && constSort);
 primBoolSort = boolSort->GetFirstElem()->GetSort();
      if (primBoolSort == s_bool)   primConstSort = s_const;
 else if (primBoolSort == s_pred)   primConstSort = s_statefct;
 else if (primBoolSort == s_action) primConstSort = s_transfct;
 else primConstSort = s_joker;
 delete boolSort;
 return MaxMatch(new Sort(primConstSort),constSort);
}

Sort* MaxBoolMatchOfMixedSorts(Sort *boolSort, Sort *constSort)
 {
 Sort_Type primConstSort, primBoolSort;

 assert(boolSort && constSort);
 primBoolSort = boolSort->GetFirstElem()->GetSort();
 if (primBoolSort == s_temporal)
   {
   delete constSort;
   return boolSort;
   }
 primConstSort = constSort->GetFirstElem()->GetSort();
      if (primConstSort == s_const)    primBoolSort = s_bool;
 else if (primConstSort == s_statefct) primBoolSort = s_pred;
 else if (primConstSort == s_transfct) primBoolSort = s_action;
 else primBoolSort = s_joker;
 delete constSort;
 return MaxMatch(new Sort(primBoolSort),boolSort);
}
