/*----------------------------------------------------------------------*
 * SymList.h
 *
 * Michael Geikowski, 14.6.93
 *
 * Definiert die Klasse SymList sowie deren Hilfsklassen
 * SymListElem und id_info.
 * Symbollisten (also Objekte der Klasse SymList) dienen dazu, Symbole, die
 * vom Scanner geliefert werden, in einer Liste zu speichern, so dass sie
 * durch den Parser mit Informationen versehen werden koennen und spaeter als
 * Beschreibung des Quelltextes in der Datenbasis abgelegt werden koennen.
 * Jedes Element einer Symbolliste --- SymListElem --- enthaelt die
 * Informationen genau eines Symbols. Handelt es sich bei dem Symbol um einen
 * Bezeichner, so werden zusaetzliche Informationen in einem Objekt der Klasse
 * id_info abgelegt, das dem SymListElem zugeordnet wird.
 *----------------------------------------------------------------------*/
/* Olaf Meier,16.6.94
 *
 * Ein Objekt der Klasse id_info besitzt nun zusaetzlich Typinformationen,
 * sowie einen Verweis auf den Bezeichner mit der aktuellen Typinformation,
 * das sich die Typen eines Bezeichners durch Inclusion mit Parameteraustausch
 * veraendern koennen.
 *----------------------------------------------------------------------*/ 






#ifndef SYMLIST_H
#define SYMLIST_H

#include <assert.h>

#include "list.h"
#include "sort.h"
#include "tlatype.h"  // Klasse TypeComp zur Darstellung von Typinformationen 



/*------Bezeichner Typen------------------------------------------------*/

enum normalID_Type
 {
 nidt_normal,
 nidt_moduleName,	nidt_imported,		nidt_included,
 nidt_expOperator,	nidt_expTheorem,
 nidt_qualifier,
 nidt_formalPar, 	nidt_formalglobPar,	nidt_substPar,
 nidt_operator, 	nidt_fctOperator,
 nidt_assumption,	nidt_theorem,
 nidt_boundVar, 	nidt_rec_comp
 };



/*------Modul Typen-----------------------------------------------------*/

enum Module_Type {mt_const = 'C', mt_temporal = 'T' };

class SymList;
class SymListElem;




class id_info
//===========
	// Speichert Informationen zu einem Bezeichner.
 {
 friend class SymListElem;
 friend class SymList;

 public:

   id_info(char* l = 0, normalID_Type t = nidt_normal, char p = 0);
 //----------------------------------------------------------------
	// Neues Objekt id_info mit Lexem `l', Typ `t' und Prime Flag `p'.
	// Zeiger werden auf NULL, qualifier auf leere Liste initialisiert.

   ~id_info();
	// Loescht gegebenenfalls Lexem, Sorte und Qualifiziererliste.

 private:

  char*           lexem;	  // Name des Bezeichners
  char		  primed;	  // True, wenn Bezeichner `primed'
  normalID_Type   nid_type;	  // Typ
  Sort*		  the_Sort;	  // Sorte
  SymList*	  qualifier;	  // Qualifiziererliste
  TypeComp*       TypeInfo;       // Typinformationen ->om
  SymListElem*    TypeDefinition; // Bezeichner mit aktueller Typinfo->om

  char*		  definingModule; // Name des Moduls, in dem die Definition des
				  // Bezeichners steht.
  SymListElem* definedAt;	  // Definition des Bezeichners oder NULL,
				  // wenn `this' selbst die Definition ist
  SymListElem* nextUse;		  // Naechste Verwendung des Bezeichners
  SymListElem* prevUse;		  // Letzte Verwendung des Bezeichners
 };

class SymListElem : public ListElem
//=================================
	// Speichert Informationen zu einem Symbol
 {
 friend class SymList;

 public:

   long pos;	// Position des Lexems
   int  leng;	// Laenge des Lexems
   int tokenID; // TokenID des Lexems ->om
   int line;     // Zeile im Spezifikationsmodul
   char* tokenName; // Name des Tokens -> fuer die Erzeugung des Parsetrees

   SymListElem(long p, int l,int lin, int tid = 0, id_info *id_p = 0)
 //---------------------------------------------
	:pos(p), leng(l),line(lin) , tokenID(tid), id_info_ptr(id_p), tokenName(NULL)
	{}

   ~SymListElem()
 //--------------
	{ if(tokenName) delete tokenName; 
          if (id_info_ptr) delete id_info_ptr; }

   char* Getlexem()
 //----------------
	{ assert(id_info_ptr);
	  return id_info_ptr->lexem; }

   void SetSort(Sort *s)
 //---------------------
	{ assert(id_info_ptr);
	  id_info_ptr->the_Sort = s; }

   void SetSort(SortListElem *e)
 //-----------------------------
	{ assert(id_info_ptr);
	  id_info_ptr->the_Sort = new Sort(e); }

   void SetTypeInfo(TypeComp* type)  // ->om
 //--------------------------------
	{ assert(id_info_ptr);
	  id_info_ptr->TypeInfo = type; }

   TypeComp* GetTypeInfo()
 //--------------------------------  // ->om
	{ assert(id_info_ptr);
	  return id_info_ptr->TypeInfo;  }

   SymListElem* GetTypeDefinition()
 //--------------------------------- // ->om
	{ assert(id_info_ptr);
          return id_info_ptr->TypeDefinition; }


   void SetTypeDefinition(SymListElem* def)
 //--------------------------------- // ->om
       {assert(id_info_ptr);
        id_info_ptr->TypeDefinition = def;
       }

   Sort* GetSort()
 //---------------
	{ assert(id_info_ptr);
	  return id_info_ptr->the_Sort; }

   void Set_nidt(normalID_Type t)
 //------------------------------
	{ assert(id_info_ptr);
	  id_info_ptr->nid_type = t; }

   normalID_Type Get_nidt()
 //------------------------
	{ assert(id_info_ptr);
	  return id_info_ptr->nid_type; }

   int is_primed()
 //---------------
	{ assert(id_info_ptr);
	  return id_info_ptr->primed; }

   void SetPrime(char p)
 //---------------------
	{ assert(id_info_ptr);
	  id_info_ptr->primed = p; }

   void SetDefinedAt(SymListElem *e);
 //----------------------------------
	// Setze Bezeichner `e' als Definition des Bezeichners `this' fest.
	// Aktualisiere dabei die Liste der Verwendungen des Bezeichners
	// `this'.

   void SetDefiningModule(char *defMod)
 //------------------------------------
	{ assert(id_info_ptr);
	  id_info_ptr->definingModule = defMod; }

   char* GetDefiningModule()
 //-------------------------
	{ assert(id_info_ptr);
	  return id_info_ptr->definingModule; }

   SymListElem* copy();
 //-----------------------------
	// Erzeugt eine Kopie des Objektes.

   void AddQualifier(SymListElem *e);
 //----------------------------------
	// Haenge einen neuen Qualifier vorne an Qualifiziererliste an.

   void SetQualifier(SymList *s);
 //------------------------------
	// Setze Qualifiziererliste.

   SymList* GetQualifier();
 //------------------------
	// Liefere Qualifiziererliste.

   int lexemcmp(SymListElem *e);
 //-----------------------------
	// Vergleiche lexem mit lexem von e. Ergibt true bei Gleichheit

 private:

   void WriteDest(char* s1, char* s2, char* s3, char* DestModName = 0);
 //--------------------------------------------------------------------
	// Hilfsfunktion fuer SymList::WriteLinkFile()

   id_info 	  *id_info_ptr;
 };

class SymList : public List
//============================
 {
 public:

   SymList(SymListElem *e = 0)
 //---------------------------
	: List(e)
	{}

   ~SymList()
 //----------
	{}

   SymListElem* GetFirstElem();
 //----------------------------
 	// Starte Durchlauf und liefere erstes Element oder NULL bei leerer
        // Liste

   SymListElem* GetNextElem();
 //---------------------------
 	// Liefere naechstes Element des aktuellen Durchlaufs oder NULL fuer
        // Ende des Durchlaufs.
 
   SymListElem* Lookup(SymListElem *e, SymList* qual);
 //---------------------------------------------------
 	// Liefere ersten Bezeichner der Liste, der
        // den gleichen Namen wie der Bezeichner e kodierte hat und
        // dessen Qualifiziererliste mit der durch `qual' gegebenen
        // uebereinstimmt. Alle Symbole der Liste sowie das Symbol e muessen
        // Bezeichner sein.




   void SubstPar_with_Type(SymListElem* par, TypeComp* typ); //->om
 //---------------------------------------------------------
      // Durchlaufe Bezeichnerliste und durchsuche vorhandene
      // Typinformationen nach dem Typ 'par' und ersetze
      // ihn durch den neuen Typ 'typ'

   void WriteLinkFile(char* filename);
 //-----------------------------------
	// Schreibe Informationen ueber die Bezeichner einer Liste von Symbolen
        // unter dem Namen `filename' als Link-Datei in die Datenbasis.


 };

#endif
