
/*-----------------------------------------------------------------------------------------
 * TLAtype.h
 *
 * Olaf Meier, 16.6.94
 * 
 * Definiert die Klasse TypeComp sowie dessen Hilfsroutinen
 * Ein Objekt der Klasse TypeComp besteht aus einem der nachfolgend
 * aufgefuehrten Typflags (= 'TLA-Typ'):
 *
 * ANY_type          : repraesentiert beliebigen Typ und ist zu jedem anderen Typ kompatible
 * UNKNOWN_type      : repraesentiert unbekannten  oder unkorrekten Typ
 * NATNUM_type       : repraesentiert den Typ 'beliebige natuerliche  Zahl'
 * STRING_type       : repraesentiert den Typ 'beliebige Zeichenkette'
 * REAL_type         : repraesentiert den Typ 'Menge von beliebigen reellen Zahlen' (wird nicht genutzt)
 * EMPTYSET_type     : repraesentiert den Typ 'leere Menge' und ist zu jedem anderen Mengentyp kompatible
 * ID_type           : repraesentiert einen bis dahin noch unbekannten Typ, der erst spaeteren
 *                     Zeitpunkt ermittelt werden kann (Typen globaler Parameter)
 * OPERATOR_type     : repraesentiert eine Art 'Prozedurtyp' mit Parametertypen und RueckgabeTyp
 * ARRAY_type        : repraesentiert einen 'Abbildungstyp', der Definitionsbereichstypen
 *                     auf Bildbereichstypen abbildet.
 * TUPLE_type        : repraesentiert einen 'Vektortyp' mit einer bestimmten Anzahl von Komponententypen
 * ENUMSET_type      : repraesentiert eine Menge mit einen Zahl von Type
 * SET_type          : repraesentiert den Typ Menge von unbestimmten Typ
 * SUBSET_type       : repraesentiert den Typ Potenzmenge mit unbestimmter Grundmenge
 * UNION_type        : repraesentiert eine Art Verknuefungstyp,  der verschiedene Typen
 *                     vereinigen kann
 * RECORD_type       : repraesentiert eine Struktur mit verschieden Komponentennamen bestimmten Typs
 * BOOLEAN_type        : repraesentiert den booleschen Typ 'true' oder 'false'
 * JOKER_type        : repraesentiert einen Typ fuer Systemvariablen der Systemformel,
 *                     der spaeter erst konkret angegeben werden kann.
 *
 * Jeder Typinformation ist baumartig aufgebaut und kann bis zu 3 (comp1...comp3) Nachfolge
 * glieder besitzen. Da der Typ globaler Parameter u.U. erst zum spaeteren Zeitpunkt
 * bekannt ist (declare after use) beinhaltet 'ID_Type' einen Verweiss auf einen Teilbaum
 * des Ableitungsbaumes, um eine notwendige Traversierung nachholen zu koennen.
 * 'compname' wird fuer den Typ 'RECORD_type' benoetigt und vermerkt die Komponentennamen.
 * 'parameter' vermerkt den Namen von globalen Parameteren, dessen Typangaben durch
 * andere Typen (bei Includes mit Parameter-Substitutionen) moeglicherweise ersetzt
 * werden.
 *-----------------------------------------------------------------------------------------*/



#ifndef TLATYPE_H
#define TLATYPE_H

#include <assert.h>

// Die tTLA-Typen:

enum TLA_Type
 { ANY_type, BOOLEAN_type,
 UNKNOWN_type, NATNUM_type, STRING_type, REAL_type,
 EMPTYSET_type, ID_type, OPERATOR_type, ARRAY_type, TUPLE_type, ENUMSET_type,
 SET_type, SUBSET_type, UNION_type, RECORD_type, JOKER_type };





class TypeComp
//=============
     // Speichert Typinformationen zu einem Bezeichner.
{
 public:

    TypeComp(TLA_Type t = UNKNOWN_type)
     : type(t), compname(NULL) , comp1(NULL), comp2(NULL), comp3(NULL), ID_Type(NULL), parameter(NULL)
     {};

   ~TypeComp()
    {
      if(comp1) 
       {
        delete comp1;
        comp1 = NULL;
       };
      if(comp2) 
       {
         delete comp2;
         comp2 = NULL;
       };
      if(comp3)
       {
         delete comp3;
         comp3 = NULL;
       };

    };

    void SetType(TLA_Type typ)
     {
       type = typ;
     };

    void SetCompname(char* name)
     {
       compname = name;
     };

    void SetParname(char* name)
     {
       parameter = name;
     };
    
    void SetComp1(TypeComp* tc1)
     {
       comp1 = tc1;
     };

    void SetComp2(TypeComp* tc2)
     {
       comp2 = tc2;
     };


    void SetComp3(TypeComp* tc3)
     {
       comp3 = tc3;
     };

    void SetID_Type(void* ptr)
     { 
       ID_Type = ptr;
     };

    TLA_Type GetType()
     {
       return type;
     };

    TypeComp* GetComp1()
     {
       return comp1;
     };


    TypeComp* GetComp2()
     {
       return comp2;
     };


    TypeComp* GetComp3()
     {
       return comp3;
     };

    TypeComp* GetNextType()
     {
       return comp3;
     };

    void* GetID_Type()
     {
      assert(ID_Type);
      return ID_Type;
     };

    char* GetCompname()
     {
     
      return compname;
     };

    char* GetParname()
     {
      return parameter;
     };

// GetTypeofTupleComp() gibt den Typ der `number'ten Vektorkomponente
// zurueck.
   
   TypeComp* GetTypeofTupleComp(int number);


// GetTypeofRecComp() gibt den Typ der Recordkomponente mit dem Namen 'comp'
// zurueck.

   TypeComp* GetTypeofRecComp(char* comp);

// Copy() kopiert die Typinformation, wobei 'comp3' (fuerParametertypenliste) 
// nicht beruecksichtigt wird.
 
   TypeComp* Copy();
  
// CopyRest() kopiert die komplette Typinformation.

   TypeComp* CopyRest();

// PrintType_nl() stellt die Typen im  ASCII-Format, gefolgt von mit einem 'newline', dar.
// (benutzt  PrintType())

   void PrintType_nl();

//  PrintType() stellt die Typen im  ASCII-Format dar.

   void PrintType();

// PrintParameters() stellt die Parametertypenliste im  ASCII-Format dar.
 
   void PrintParameters();

//  PrintTupleComps() stellt die Vektorkomponenten im  ASCII-Format dar.

   void PrintTupleComps(int n);

// PrintRecordComps() stellt die Strukturkomponenten im  ASCII-Format dar.

   void PrintRecordComps();

// TraverseTypeInfo ueberprueft Typinformation 'this', ob er einen Typ vom globalen Parameter
// mit dem Namen 'par' besitzt. Falls ja, wird dieser Typ durch den Typ 'typ' substituiert.

   void TraverseTypeInfo(char* par, TypeComp* typ);

   private:

      TLA_Type type; // Typ-Flag

      char* parameter; // Parametername falls Typ vom globalen Parameter uebernommen   

      char* compname;  // Struktur-Komponentennamen

      TypeComp* comp1; // Verweis auf Parametertypenliste

      TypeComp* comp2; // Verweis auf Rueckgabetypenliste
                       
      TypeComp* comp3; // Verweis auf naechten Parametertyp

      void* ID_Type;  // Verweis auf Teilbaum den Parsetrees ( ParseExpr) oder auf Bezeicher (SymListElem)
  };

/*-------------------------------------------------------------------------------------
 * Non-Member Funktionen
 *------------------------------------------------------------------------------------*/


//  GetLastType() gibt -- ausgehend von Typglied 'typ' --  den letzten Typ der Typkette zurueck

TypeComp* GetLastType(TypeComp* typ);


//  GetElementofType() gibt beim 'Mengen-Typ "type"'  die Mengenelemente zurueck, im Fehlerfall NULL .

TypeComp* GetElementofType(TypeComp* type);


// GetARRAY() liefert 'type' wieder zurueck, falls es sich um einen ARRAY Typ handelt, NULL sonst.

TypeComp* GetARRAY(TypeComp* type);

// GetTUPLE()  liefert 'type' wieder zurueck, falls es sich um einen Vektortyp handelt, sonst NULL.
// Wird zusaetzlich 'num' angegeben, so wird der Typ der 'num'ten Komponente zurueckgegeben, im 
// Fehlerfall NULL. 
TypeComp* GetTUPLE(TypeComp* type, int num = 0);


//  GetRECORD()  liefert 'type' wieder zurueck, falls es sich um einen Strukturtyp handelt, sonst NULL.
// Wird zusaetzlich 'comp' angegeben, so wird der Typ der Komponente 'comp' zurueckgegeben,im 
// Fehlerfall NULL.

TypeComp* GetRECORD(TypeComp* type,char* comp = NULL);

#endif
