/*-------------------------------------------------------------------------*
 * tppmain.c
 *
 * Michael Geikowski, 16.6.93
 *
 * TLA+ Praeprozessor Steuerung
 *-------------------------------------------------------------------------*/
/* Ergaenzung: Olaf Meier, 10.9.94
 * Aufbereitung von Informationen fuer die spaetere semantische Konzepterfassung.
 *-------------------------------------------------------------------------------*/

#include <new.h>
#include <stdlib.h>
#include <stdio.h>

#include "adjlistgnu.h"
#include "higraphgnu.h"
#include "layout.h"
#include "crossref.h"
#include "dbinterface.h"
#include "parsetree.h"
#include "symlist.h"
#include "semconcept.h"

// Ersatz des Headers fuer Parser1

extern AdjList*  the_AdjList;	// Adjazenzliste des Modulgraphen
extern NodeList* NodeQueue;	// Warteschlange noch zu untersuchender Knoten
extern Node*     this_module;	// aktueller Knoten

extern void parse_pass1 (char* filename);

// Ersatz des Headers fuer Parser2

extern CrossRef* crossref;	// Querverweisliste

extern int       yydebug;

extern void parse_pass2 (char* suffixedfilename, char* filename);

// Ersatz des Headers fuer den Scanner

extern int       debug_ips;
extern int       lex_list;

// Externe Variable fuer getopt()

extern int       optind;



// Externe Variablen fuer semantische Konzepterfassung

extern ParseExpr* pe_semantic_concept;    // Expandierter Parsetree der Systemformel
extern ParseExpr* pe_variables;           // Parsetree aller benutzten Variablen
AdjList* the_Variables;

static void usage()
 {
 fprintf(stderr, "usage: tpp [-ntil] <spec-file>\n");
 fprintf(stderr, " -n : no standalone modus\n");
 fprintf(stderr, " -t : rule trace (parser)\n");
 fprintf(stderr, " -i : indentation debug (scanner)\n");
 fprintf(stderr, " -l : list spec-file (scanner)\n");
 fflush(stderr);
 }

static void OutOfMemory()
 { 
 fprintf(stderr,"Sorry, out of memory\n");
 fflush(stderr);
 exit(-1);
 }

int main (int argc, char** argv)
 {
 char     arg;
 char    *filename,*p,*dot,*suffix,*name;
 HiGraph *var_hg =NULL,*hg =NULL;
 
 Node*    n;
 ParseExpr* tree;
 SymList *Symbollist = new SymList; 
 the_Variables = new AdjList;

 the_AdjList = new AdjList;	// Init globaler Variablen
 NodeQueue   = new NodeList;
 this_module = 0;
 crossref    = new CrossRef;
 yydebug     = 0;
 debug_ips   = 0;
 lex_list    = 0;
 standalone  = 1;
 set_new_handler(OutOfMemory);	// Error Handler init
 while ((arg = getopt(argc, argv, "ntil")) != EOF)	// Optionen
   {
   switch (arg)
     {
     case 'i': debug_ips = 1;
               break;
     case 't': yydebug = 1;
               break;
     case 'l': lex_list = 1;
     	       break;
     case 'n': standalone = 0;
     	       break;
     case '?': usage();
               exit(-1);
     }
   }
 if (optind != argc-1)
   {
   usage();
   exit(-1);
   }
 
 // Suffix von Filenamen trennen

 p   = argv[optind];
 dot = 0;
 while ( *p != '\0' )
   {
   if (*p=='.') { dot=p; }
   p++;
   }
 if (dot)
   {
   *dot = '\0';
   dot++;
   suffix = new char[strlen(dot)+2];
   sprintf(suffix,".%s",dot);
   }
 else
   {
   suffix = new char[1];
   suffix[0] = '\0';
   }
   
 // Modul Hierarchie feststellen mittels BFS
 
 fprintf(stderr,"reading module hierarchie...\n");
 fflush(stderr);
 name = argv[optind];
 this_module = the_AdjList->AddNode(name);
 while (name)
   {
   fprintf(stderr,"...found '%s'\n",name);
   fflush(stderr);
   filename = new (char[strlen(name)+strlen(suffix)+1]);
   sprintf(filename,"%s%s",name,suffix);
   parse_pass1(filename);
   delete filename;
   this_module = (Node*) NodeQueue->GetFirstElem();
   if (this_module)
     {
     this_module->SetCounter(1);
     name = this_module->GetName();
     NodeQueue->DelFirstElem();
     }
   else
     {
     name = 0;
     }
   }

 // Graph Layout

 fprintf(stderr,"module graph layout\n");
 fflush(stderr);
 hg = Layout(the_AdjList,argv[optind]);
 if (!hg)
   {
   fprintf(stderr,"module graph is cyclic\n");
   fprintf(stderr,"-> cannot read this specification\n");
   fflush(stderr);
   delete crossref;
   delete the_AdjList;
   delete NodeQueue;
   delete suffix;
   delete the_Variables;
   exit(-1);
   }
 
 // Spezifikation parsen. Modul Hierarchie bottom up lesen, um Importe
 // und Includes behandeln zu koennen
 
 fprintf(stderr,"reading modules...\n");
 fflush(stderr);
 for (int Level=hg->GetNrOfLevels()-1; Level >= 0; Level--)
   {
   n = hg->FirstNodeOfLevel(Level);
   while(n)
     {
     if (n->GetNodeType() != 'D')
       {
       name = n->GetName();
       fprintf(stderr,"...'%s'\n",name);
       fflush(stderr);
       filename = new (char[strlen(name)+strlen(suffix)+1]);
       sprintf(filename,"%s%s",name,suffix);
       parse_pass2(filename, name);

       delete filename;
       
       }
     n = hg->NextNodeOfLevel(Level);
     }
   }

 // Vorbereitung der Semantische Konzepterfassung

   tree = crossref->GetParseTree(name);  // ableitungsbaum des Topmoduls
   
   if(tree) 
    {   
      fprintf(stderr,"searching for system formula...\n");  

      
     // Suche nach der Systemformel und
     // vollstaendige Expansion derselben.
      ExamineSystemFormula(tree,tree,Symbollist); 
      if(pe_variables)     
         { 
           fprintf(stderr,"...found system formula.\n");
           // Ermittlung der verschieden Aktionen und den in ihr
           // vorkommenden Systemvariablen.
           CreateSemanticConcept(pe_variables,pe_semantic_concept,the_Variables);
           // Erstellung des Variablenbaum-Layouts
           var_hg = VarLayout(the_Variables,argv[optind]);
           delete var_hg;
         }
      else
        {
          fprintf(stderr,"...can't find system formula.\n");
        } 
    }
 // Querverweise schreiben.
 
 fprintf(stderr,"generating links\n");
 fflush(stderr);
 crossref->WriteLinkFiles();

 // leere Lesezeichen Datei erzeugen

 name = argv[optind];			
 if (! OpenBookmarkFile(name,"w"))
   {
   fprintf(stderr,"cannot create bookmarkfile for spec %s\n",name);
   fflush(stderr);
   }
 else
   {
   CloseBookmarkFile();
   }

 delete crossref;
 delete the_AdjList;
 delete hg;

 delete NodeQueue;
 delete suffix;
 delete Symbollist;
 delete the_Variables;
 exit( 0 );
}
