MODULE SeqDiscoveryActions

  IMPORT Graph,NeedfulThings;

  PARAMETERS Initiator : Node CONSTANT;
             Active    : SET [ Node -> Boolean ];
             ToScan    : SET [ Node -> SUBSET(Node) ];
             MBox      : SET [ Node -> Node :cup: { 0 } ];
             Master    : SET [ Node -> Node ];
             WaitingOf : SET [ Node -> Node :cup: { 0 } ];

  CONSTANT ASSUMPTIONS

    NodeNames == ~ 0 :in: Node
    ;

  PREDICATES

    Init ==
      /\ Active[Initiator] = "true"
      /\ :A: v :in: Node :: v # Initiator => Active[v] = "false"
      /\ :A: v :in: Node :: /\ ToScan[v]    = Neighbours(v)
                            /\ MBox[v]      = 0
                            /\ WaitingOf[v] = 0
      ;

  ACTIONS

    SendScan (v,w:Node) ==
      /\ Active[v] = "true"
      /\ w :in: ToScan[v]
      /\ Active[v]'    = "false"
      /\ ToScan[v]'    = ToScan[v] \ { w }
      /\ MBox[w]'      = v
      /\ WaitingOf[v]' = w
      /\ :A: u :in: Node :: /\ u # v => /\ Active[u]'    = Active[u]
                                        /\ ToScan[u]'    = ToScan[u]
                                        /\ WaitingOf[u]' = WaitingOf[u]
                            /\ u # w => MBox[u]' = MBox[u]
    ;

    ReceiveFirstScan (v:Node) ==
      /\ WaitingOf[v] = 0
      /\ MBox[v]      # 0
      /\ Active[v]'   = "true"
      /\ Master[v]'   = MBox[v]
      /\ MBox[v]'     = 0
      /\ ToScan[v]'   = ToScan[v] \ { MBox[v] }
      /\ :A: w :in: Node :: w # v => /\ Active[w]' = Active[w]
                                     /\ Master[w]' = Master[w]
                                     /\ MBox[w]'   = MBox[w]
                                     /\ ToScan[w]' = ToScan[w]
      /\ Unchanged(WaitingOf)
    ;

    (* Verbesserung ReceiveSucc :         *)
    (*  /\ ToScan[v]' = ToScan[v] \ { w } *)

    ReceiveSucc (v,w:Node) ==
      /\ WaitingOf[v] # 0
      /\ MBox[v]      = w
      /\ WaitingOf[v] # w
      /\ MBox[v]'     = 0
      /\ MBox[w]'     = v
      /\ :A: u :in: Node :: (u # v /\ u # w ) => MBox[u]' = MBox[u]
      /\ Unchanged(Active,ToScan,Master,WaitingOf)
    ;

    ReceiveAnswer (v:Node) ==
      /\ WaitingOf[v] = MBox[v]
      /\ MBox[v]      # 0
      /\ MBox[v]'     = 0
      /\ Active[v]'   = "true"
      /\ :A: w :in: Node :: w # v => /\ Active[w]' = Active[w]
                                     /\ MBox[w]'   = MBox[w]
      /\ Unchanged(ToScan,Master,WaitingOf)
    ;

    NonInitiatorTermination (v:Node) ==
      /\ v                   # Initiator
      /\ Active[v]           = "true"
      /\ ToScan[v]           = {}
      /\ Active[v]'          = "false"
      /\ MBox[ Master [v] ]' = v
      /\ WaitingOf[v]'       = v
      /\ :A: w :in: Node :: /\ w # v => /\ Active[w]'    = Active[w]
                                        /\ WaitingOf[w]' = WaitingOf[w]
                            /\ w # Master[v] => MBox[w]' = MBox[w]
      /\ Unchanged(ToScan,Master)
    ;

    InitiatorTermination (v:Node) ==
      /\ v          = Initiator
      /\ Active[v]  = "true"
      /\ ToScan[v]  = {}
      /\ Active[v]' = "false"
      /\ :A: w :in: Node :: w # v => Active[w]' = Active[w]
      /\ Unchanged(ToScan,MBox,Master,WaitingOf)
    ;

END
